; File:  processparticle.pro
; Author:  Ted Fisher, vaguely based on the image processing routines from "simplevars.dll", written by Chris Wieland
; Purpose:  2D-S, HVPS, 3V-CPI image processing
;   Merged all versions of processparticle procedure, with system variable flag for 3V-CPI data format (!flag_3VCPI)
;   Any changes in 2D-S, HVPS, or 3V-CPI Views should be propagated to all other views.




; 2.10 TAF 2016-03-28 added mask bit filling to non-compressed images
; 2.9  TAF 2016-03-24 added checking for identical points in getangle, which were causing repeated floating point errors -- likely should check calling
;                     getangle, and just avoid calling it if the points don't make sense
; 2.8  TAF 2016-03-21 additional error handling for irregular image data
; 2.7  TAF 2016-03-18 fixed some missing values L5 and midpoint not being assigned, finished recompression
; 2.6  TAF 2016-02-22 added image recompression for generating SORTED files with mask-bit filled images
; 2.5  TAF 2016-01-27 added mask bit filling -- have 1 pixel, but need to implement multi-pixel, also changed to a vector-based image decompression
; 2.4  TAF 2014-10-14 csp was being calculated incorrectly, largest_satellite_area was already in units of microns^2, but was being multiplied by
;                     microns_per_pixel * microns_per_pixel again
; 2.3  TAF 2014-10-13 fixed bug with !flag_3VCPI not being defined in 2D-S data with incorrect 0x7FFF flag
; 2.2  TAF 2014-10-09 updated processparticle handling for single pixel case, number of satellites from 0 to 1, and largest_satellite_area set to 1
; 2.1  TAF 2014-10-08 modified to output largest_satellite_area, and to use largest_satellite_area in the csp calculation
; 2.0  TAF 2014-10-03 initial version number for separated processparticle.pro




pro index2xy, m, index, x, y
    ; given a linear index into a 2D rectangular m x n matrix
    ; return the x and y indices of the element in the matrix

    x = long(index mod m)
    y = long(index / m)
end ; index2xy



function modpi, angle
    ; returns the angle mod PI, (in radians) 0 to PI (0 to 180 degrees)

    ; perhaps in IDL the mod operator works on floats, might be able to update/abandon this function
    ; in c++ could try using fmod for this functionality

    ; not array safe method
    ;    if(angle lt 0.) then begin
    ;        while(angle lt 0.) do begin
    ;            angle += !PI
    ;        endwhile
    ;    endif else if(angle gt !PI) then begin
    ;        while(angle ge !PI) do begin
    ;            angle -= !PI
    ;        endwhile
    ;    endif

    ; array safe method
    done = 0

    while(~done) do begin
        index = where(angle lt 0., count)

        if(count gt 0) then begin
            angle[index] += !PI
        endif else begin
            done = 1
        endelse
    endwhile

    done = 0

    while(~done) do begin
        index = where(angle gt !PI, count)

        if(count gt 0) then begin
            angle[index] -= !PI
        endif else begin
            done = 1
        endelse
    endwhile

    return, angle
end ; modpi



function getangle, p1x, p1y, p2x, p2y, p3x, p3y
    ; getangle() determine angle at center point P2 formed by lines P1-P2 and P2-P3
    ;   return in degrees

    ;    P1    P3
    ;     |   /
    ;     |  /
    ;     | /
    ;     |/
    ;    P2

    ; test for identical points, which cause floating point errors below (which slows down processing and is wrong, since it is morally wrong to divide by zero)
    if((p1x eq p2x and p1y eq p2y) or $
        (p1x eq p3x and p1y eq p3y) or $
        (p3x eq p2x and p3y eq p2y)) then begin
        return, 0.
    endif

    centerx = p2x
    centery = p2y

    p1c = sqrt(((centerx - p1x) * (centerx - p1x)) +((centery - p1y) * (centery - p1y)))
    p2c = sqrt(((centerx - p3x) * (centerx - p3x)) +((centery - p3y) * (centery - p3y)))

    p12 = sqrt(((p1x - p3x) * (p1x - p3x)) +((p1y - p3y) * (p1y - p3y)))

    r1 = (p2c*p2c + p1c*p1c - p12*p12)/ (2.0 * p2c * p1c)

    if(r1 gt 1.00) then begin
        r1 = 1.0
    endif

    if(r1 lt -1.00) then begin
        r1 = -1.0
    endif

    xx = acos(r1) / !dtor

    return, xx
end ; getangle



;function circle_from_three_points, p1, p2, p3, center, radius
;    ; Circle Through Three Points
;    ; p1 and p3 should always be the two edge points
;    ; p2 should be the point between the two edge points.
;    ; Otherwise you try to get the slope of p1:p3 and will
;    ; produce problems in the routine since it forms a vertical line

;    ;   p1--->      .
;    ;               ..
;    ;               ...
;    ;               ....    <--- p2
;    ;               ...
;    ;               ..
;    ;   p3--->      .

;    ; See the following for an explanation of how to
;    ; determine the center and radius of a circle when
;    ; provided with three points on the circle.
;    ; http://paulbourke.net/geometry/circlefrom3/


;    x1 = p1[0]
;    y1 = p1[1]
;    x2 = p2[0]
;    y2 = p2[1]
;    x3 = p3[0]
;    y3 = p3[1]

;    yDelta_a = y2 - y1
;    xDelta_a = x2 - x1
;    yDelta_b = y3 - y2
;    xDelta_b = x3 - x2

;    if(abs(xDelta_a) le 0.000000001 and abs(yDelta_b) le 0.000000001) then begin
;        center[0] = 0.5 * (x2 + x3)
;        center[1] = 0.5 * (y1 + y2)

;        radius = (((center[0] - x1) * (center[0] - x1)) + ((center[1] - y1) * (center[1] - y1)))^.5

;        return, -1
;    endif

;    aSlope = yDelta_a / xDelta_a
;    bSlope = yDelta_b / xDelta_b

;    if(aSlope == 0.0 or bSlope == 0.0) then begin
;        return, -1 ; a slope of 0 should not be possible if the three points are to form a circle
;    endif

;    if(abs(aSlope - bSlope) <= 0.000000001) then begin
;        ; checking whether the given points are colinear.
;        ; TRACE("The three pts are colinear\n")
;        return, -1
;    endif

;    ; calc center
;    center[0] = (aSlope * bSlope * (y1 - y3) + bSlope * (x1 + x2) - aSlope * (x2 + x3)) / (2. * (bSlope-aSlope))
;    center[1] = -1. * (center[0] - (x1 + x2) / 2.) / aSlope + (y1 + y2) / 2.

;    *radius = (((center[0] - x1) * (center[0] - x1)) + ((center[1] - y1) * (center[1] - y1)))^.5

;    return, 1
;end ; circle_from_three_points



PRO ProcessParticle, image_data, image, WordCount, Slices, Sums, L2, area, Midpoint, Edges, Err,$
    L3, L4, L5, PC3x, PC3y, PC4x, PC4y, A3, Roundness1, Roundness2, Roundness3, L1r, maxLength, maxWidth, num_perimeter_points, $
    crystal_type, perimeterum, csp, $
    ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
    perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, num_satellites=num_satellites, HorV=HorV, $
    largest_satellite_area=largest_satellite_area, mask=mask, recompressed_image=recompressed_image, new_wordcount=new_wordcount



if(HorV eq 'H') then begin
    microns_per_pixel = !MICRONS_PER_PIXELH
endif else begin
    microns_per_pixel = !MICRONS_PER_PIXELV
endelse

; TAF found an odd size for an image array, which should always be even, putting in a ;stop
; the odd size was caused by a bad copy into my test program, and should not be encountered
; in normal operation.  Still, this ;stop should be left in to help debug if an odd Image is
; ever encountered.

if((size(image_data))[1] mod 2 eq 1) then begin
    print, "image_data size is odd, should be 2*WordCount long"
    help, image_data, WordCount
    print, "TAF should mark as bad particle and move on..."
    return
    ;stop
endif

if(Slices eq 0) then begin
    print, "Slices equals 0, processparticle.pro"
    help, Wordcount, Slices
    ;print, "TAF should mark as bad particle and move on..."
    return
    ;stop
endif

; for compatibility with older versions for testing
if(n_elements(mask) eq 0) then begin
    mask = uint(['FFFF'X,'FFFF'X,'FFFF'X,'FFFF'X,'FFFF'X,'FFFF'X,'FFFF'X,'FFFF'X])
endif

ml_p1x = 0
ml_p1y = 0
ml_p2x = 0
ml_p2y = 0
mw_p1x = 0
mw_p1y = 0
mw_p2x = 0
mw_p2y = 0


;     The perimeter length is initially calculated in pixels.  This can be scaled linearly for a different pixel resolution.

;     The algorithm is based on the "Improved Algorithm for Calculating the Perimeter of Raster Polygons" by Steve Prashker
;     from http://www.geocomputation.org/1999/076/gc_076.htm

;     based on the following tables:

;       P1 P2 P3           1    2    4         0x01    0x02    0x04
;       P0  P P4           8    0   16         0x08    0x00    0x10
;       P7 P6 P5          32   64  128         0x20    0x40    0x80
;    //

neighbor_codes = [8, 1, 2, 4, 16, 32, 64, 128]

; these lengths are in pixels, multiply by 10 microns/pixel for microns for regular 2D-S
length_table = $
    [4.,     4.,     3.,     1.414,  4.,     4.,     1.414,  2.828,  3.,     1.414,  2.,     2.,     3.,     1.414,  1.414,  1.414, $
     3.,     3.,     2.,     1.414,  1.414,  1.414,  2.,     1.414,  2.,     2.,     1.,     1.,     2.,     2.,     1.,     1.,    $
     4.,     4.,     3.,     1.414,  4.,     4.,     3.,     3.,     1.414,  2.828,  1.414,  1.414,  3.,     3.,     1.414,  1.414, $
     3.,     3.,     2.,     2.,     3.,     3.,     2.,     2.,     2.,     2.,     1.,     1.414,  2.,     2.,     1.,     1.,    $
     3.,     3.,     2.,     2.,     3.,     3.,     2.,     2.,     2.,     1.414,  1.,     1.,     2.,     2.,     1.,     1.414, $
     2.,     2.,     1.,     1.,     1.414,  2.,     1.,     1.414,  1.,     1.,     0.,     0.,     1.,     2.,     0.,     0.,    $
     1.414,  1.414,  2.,     2.,     3.,     3.,     2.,     2.,     2.,     1.414,  1.,     1.,     2.,     2.,     1.,     1.,    $
     1.414,  2.,     1.,     2.,     1.414,  1.414,  1.,     1.,     1.,     1.414,  0.,     0.,     1.,     1.,     0.,     0.,    $
     4.,     4.,     3.,     3.,     4.,     4.,     1.414,  3.,     3.,     3.,     2.,     2.,     3.,     3.,     2.,     2.,    $
     1.414,  3.,     1.414,  1.414,  2.828,  3.,     1.414,  1.414,  2.,     2.,     1.,     1.,     2.,     2.,     1.414,  1.,    $
     4.,     4.,     3.,     3.,     4.,     4.,     3.,     3.,     1.414,  3.,     2.,     2.,     3.,     3.,     1.414,  1.414, $
     1.414,  3.,     2.,     1.414,  3.,     3.,     2.,     1.414,  2.,     2.,     2.,     1.,     2.,     1.414,  1.,     1.,    $
     1.414,  3.,     2.,     2.,     1.414,  3.,     2.,     2.,     1.414,  1.414,  1.,     1.,     2.,     1.414,  2.,     1.,    $
     2.,     2.,     1.,     1.,     1.414,  2.,     1.,     1.,     1.,     1.,     0.,     0.,     1.414,  1.,     0.,     0.,    $
     2.828,  3.,     2.,     2.,     3.,     3.,     2.,     1.414,  1.414,  1.414,  1.414,  1.,     2.,     1.414,  1.,     1.,    $
     1.414,  2.,     1.414,  2.,     1.414,  1.414,  1.,     1.,     1.,     1.,     0.,     0.,     1.,     1.,     0.,     0.     ]

; edges, also known as F1 is used as a set of binary flags with the following values:

; 0x01 - "right edge", pixel 127 is shaded
; 0x02 - "left  edge", pixel   0 is shaded
; 0x04 - Roundness calculated for an edge particle
; 0x08 - center of centroid is on array
; 0x10 - Roundness calculated for a dual edge particle
; 0x20 - perimeter vs area criteria not met
; 0x40 - image contains satellites (shaded regions are disjoint


; decompress image
; image should be wordcount words long

image_data_size = 2L * long(wordcount)
image_size = long(!PIXELS) * long(slices)

if(image_size lt 0) then begin
    print, "image_size exceeds numerical limits error ", image_size

    return ;stop
endif

if((size(image_data))[1] ne image_data_size) then begin
    print, "image size not equal to specified wordcount, attempting to use truncated image"

    wordcount = 2L * long((size(image_data))[1])
endif

current_wordcount = 0L
;current_slices    = 0L

current_slice = -1L
current_pixel =  0L

dptr = 0L

max_shaded_bits = 0 ; L2, most shaded bits in a slice
max_slice_width = 0 ; L4, the widest single slice

; used for the entire particle, can calculate L5, the max width of the particle across all slices
first_shaded_bit = !PIXELS
 last_shaded_bit = -1

total_shaded_pixels = 0L

edges = 0

 total_bits = 0L
shaded_bits = 0L
 clear_bits = 0L

image[*] = byte(255) ; for now, at this point, 255 is clear, 0 is shaded
image2   = bytarr(!PIXELS, slices) ; used for binary image processing below

retval = 0
done   = 0

this_slice_width = 0 ; TAF had a crash where this wasn't defined
total_shaded_pixels = 0

first_shaded_bit = !PIXELS + 1
 last_shaded_bit = -1

max_shaded_bits = 0 ; L2, most shaded bits in a slice
max_slice_width = 0 ; L4, the widest single slice

; reorder the bytes in the image_data and cast into words for decompression
; NOTE: this can likely be done in one step, check setup before call to processparticle, ie in batch2d
index = indgen(wordcount)

hi_bytes = byte(image_data[2*index+1], 0, 1, wordcount)
lo_bytes = byte(image_data[2*index  ], 0, 1, wordcount)

image_words = uint([lo_bytes, hi_bytes], 0, wordcount)

; initialize both of these for the image decompression loop logic
noncompressed_slice_count = 0
        clear_slice_count = 0

; NOTE: these are potential slice_start indices -- the noncompressed 3V types will have data words that match this -- they are handled separately below
slice_start_index = where((image_words[*] and '4000'X) ne 0, slice_start_count)

if(slice_start_count lt 1) then begin
    print, 'no slice start flags in image data'

    return
endif

; slice start flag 0x7FFF indicates a noncompressed slice in 3V types, but a clear slice (should be rare or nonexistent) in 2D types
if(!flag_3VCPI) then begin
    noncompressed_slice_index = where(image_words[slice_start_index] eq '7FFF'X, noncompressed_slice_count)
endif else begin
            clear_slice_index = where(image_words[slice_start_index] eq '7FFF'X,         clear_slice_count)
endelse

full_shaded_index = where(image_words[slice_start_index] eq '4000'X, full_shaded_count) ; 0x4000 is defined as 128 shaded bits

; add one additional element to slice_start_index, for calculating the number of words in each slice
; the value is wordcount
slice_start_index = [slice_start_index, wordcount]

nk=0
fk=0
current_slice = 0
mask_bit_filled = 0 ; flag to recompress image, only needed if a masked bit has been filled
flagged = intarr(slice_start_count) ; signals that this slice is a special flagged case and is decoded differently: noncompressed, clear, or shaded




; image decompression loop
for i=0L, slice_start_count-1 do begin
    if(!flag_3VCPI) then begin
        if((noncompressed_slice_count gt 0) and (nk lt noncompressed_slice_count)) then begin
            if(slice_start_index[i] eq slice_start_index[noncompressed_slice_index[nk]]) then begin

                ; ensure the indices are within the bounds of image_words -- truncate image if not

                if(slice_start_index[noncompressed_slice_index[nk]]+1 ge wordcount) then begin
                    print, 'noncompressed_slice_index[nk]]+1 ', slice_start_index[noncompressed_slice_index[nk]]+1, ' out of wordcount', wordcount
                    stop
                endif

                if(slice_start_index[noncompressed_slice_index[nk]]+8 lt wordcount) then begin
                    slice_words = image_words[slice_start_index[noncompressed_slice_index[nk]]+1:slice_start_index[noncompressed_slice_index[nk]]+8]
                endif else begin
                    ; truncate the image to use the available words
                    slice_words = image_words[slice_start_index[noncompressed_slice_index[nk]]+1:*]
                endelse

                nk++
                flagged[i] = 1

                ; detect and flag any potential miss identified starts of slice -- the image decompression will then skip
                miss_ident_index = where((slice_words  and '4000'X) ne 0, miss_ident_count)
                if(miss_ident_count gt 0) then begin
                    flagged[i:i+miss_ident_count] = 1
                    ;stop
                endif


                end_word = 7 < (size(slice_words))[1]-1 ; if we are using a truncated image, the decompression might be fewer than 8 words -- leave the rest clear, likely discard
                for m=0, end_word do begin
                    for n=0, 15 do begin
                        bit = '0001'X and ishft(slice_words[m], -n)

                        if(~bit) then begin
                            image [m*16+n, current_slice] = 0
                            image2[m*16+n, current_slice] = 1
                        endif
                    endfor
                endfor

                shaded_index = where(image[*, current_slice] eq 0B, shaded_count)
                if(shaded_count gt 0) then begin

                    first_shaded_bit_this_slice = shaded_index[0]
                     last_shaded_bit_this_slice = shaded_index[shaded_count-1]

                    shaded_bits = shaded_count
                endif else begin
                    print, '3V noncompressed but no shaded bits?'

                    ;stop

                    first_shaded_bit_this_slice = -1
                     last_shaded_bit_this_slice = -2

                    shaded_bits = 0
                endelse

                ; mask bit filling -- make into a function and use for both compressed and non-compressed
                if(!mask_bit_filling eq 1) then begin
                    ; naive implementation copied from Realtime 2.X

                    ; currently only fills single pixel wide masks, needs to be updated to use !mask_bit_width

                    MASK_N_FIELDS = 8
                    MASK_N_HEX_CHARS = 4
                    NBITS = 4

                    for iii=0, MASK_N_FIELDS-1 do begin
                        for jjj=0, MASK_N_HEX_CHARS-1 do begin
                            for kkk=0, NBITS-1 do begin
                                current_bit = 1U and (ishft(mask[iii], -(NBITS*jjj+kkk)))

                                if(~current_bit) then begin
                                    mask_index = (NBITS*MASK_N_HEX_CHARS*iii + NBITS*(MASK_N_HEX_CHARS-jjj-1) + (NBITS-kkk-1))

                                    if(image[mask_index, current_slice] ne 0) then begin
                                        image[mask_index, current_slice] = 170 ; colorizes the mask

                                        if(mask_index eq 0) then begin ; low edge
                                            if(image[1, current_slice] eq 0) then begin
                                                image [0, current_slice] = 220
                                                image2[0, current_slice] = 1
                                                first_shaded_bit_this_slice = 0
                                                mask_bit_filled = 1
                                                total_shaded_pixels++
                                            endif
                                        endif else if(mask_index eq !PIXELS-1) then begin ; high edge
                                            if(image[!PIXELS-2, current_slice] eq 0) then begin
                                                image [!PIXELS-1, current_slice] = 220
                                                image2[!PIXELS-1, current_slice] = 1
                                                last_shaded_bit_this_slice = !PIXELS-1
                                                mask_bit_filled = 1
                                                total_shaded_pixels++
                                            endif
                                        endif else begin ; middle pixel
                                            if((image[mask_index-1, current_slice] eq 0) and (image[mask_index+1, current_slice] eq 0)) then begin
                                                image [mask_index, current_slice] = 220
                                                image2[mask_index, current_slice] = 1
                                                mask_bit_filled = 1
                                                total_shaded_pixels++
                                            endif
                                        endelse
                                    endif
                                endif
                            endfor
                        endfor
                    endfor
                endif

                if(first_shaded_bit_this_slice lt first_shaded_bit) then first_shaded_bit = first_shaded_bit_this_slice
                if( last_shaded_bit_this_slice gt  last_shaded_bit) then  last_shaded_bit =  last_shaded_bit_this_slice

                current_slice++
            endif
        endif
    endif else begin
        if((clear_slice_count gt 0) and (nk lt clear_slice_count)) then begin
            if(slice_start_index[i] eq slice_start_index[clear_slice_index[nk]]) then begin
                print, '2D type clear slices = ', clear_slice_count

                nk++
                flagged[i] = 1

                first_shaded_bit_this_slice = -1 ; Not sure how to represent -- though this will give zero to the max_slice_width >= last_shaded_bit_this_slice - first_shaded_bit_this_slice + 1
                 last_shaded_bit_this_slice = -2

                shaded_bits = 0

                current_slice++
            endif
        endif
    endelse

    if(~flagged[i] and (full_shaded_count gt 0) and (fk lt full_shaded_count)) then begin
        if(slice_start_index[i] eq slice_start_index[full_shaded_index[fk]]) then begin
            ;print, 'full shaded slices = ', full_shaded_count

            fk++
            flagged[i] = 1

            image [*, current_slice] = 0
            image2[*, current_slice] = 1

            first_shaded_bit_this_slice = 0
             last_shaded_bit_this_slice = !PIXELS-1

            first_shaded_bit            = 0
             last_shaded_bit            = !PIXELS-1

            shaded_bits = !PIXELS

            current_slice++
        endif
    endif

    if(~flagged[i]) then begin
        ; uncompresss the image
        first_shaded_bit_this_slice = !PIXELS + 1
         last_shaded_bit_this_slice = -1

        nwords = slice_start_index[i+1] - slice_start_index[i]

        slice_words = image_words[slice_start_index[i]:slice_start_index[i+1]-1]

         clear_bits =       slice_words      and '7F'X              ; '7F'X = 0111,1111
        shaded_bits = ishft(slice_words, -7) and '7F'X

        image_indices = reform(transpose([[clear_bits], [shaded_bits]]), 2*nwords)

        index = 0

        first_shaded_bit_this_slice = clear_bits[0]

        for j=0L, nwords-1 do begin
            index += clear_bits[j]

            index2 = index + shaded_bits[j]-1

            if(index le index2 and index2 lt !PIXELS) then begin
                image [index:index2, current_slice] = 0
                image2[index:index2, current_slice] = 1
            endif

            index += shaded_bits[j]
        endfor

        last_shaded_bit_this_slice = index2

        ; mask bit filling
        if(!mask_bit_filling eq 1) then begin
            ; naive implementation copied from Realtime 2.X

            ; currently only fills single pixel wide masks, needs to be updated to use !mask_bit_width

            MASK_N_FIELDS = 8
            MASK_N_HEX_CHARS = 4
            NBITS = 4

            for iii=0, MASK_N_FIELDS-1 do begin
                for jjj=0, MASK_N_HEX_CHARS-1 do begin
                    for kkk=0, NBITS-1 do begin
                        current_bit = 1U and (ishft(mask[iii], -(NBITS*jjj+kkk)))

                        if(~current_bit) then begin
                            mask_index = (NBITS*MASK_N_HEX_CHARS*iii + NBITS*(MASK_N_HEX_CHARS-jjj-1) + (NBITS-kkk-1))

                            if(image[mask_index, current_slice] ne 0) then begin
                                image[mask_index, current_slice] = 170 ; colorizes the mask, however interferes with recompression -- need to resolve, maybe a third image?

                                if(mask_index eq 0) then begin ; low edge
                                    if(image[1, current_slice] eq 0) then begin
                                        image [0, current_slice] = 220
                                        image2[0, current_slice] = 1
                                        first_shaded_bit_this_slice = 0
                                        mask_bit_filled = 1
                                        total_shaded_pixels++
                                    endif
                                endif else if(mask_index eq !PIXELS-1) then begin ; high edge
                                    if(image[!PIXELS-2, current_slice] eq 0) then begin
                                        image [!PIXELS-1, current_slice] = 220
                                        image2[!PIXELS-1, current_slice] = 1
                                        last_shaded_bit_this_slice = !PIXELS-1
                                        mask_bit_filled = 1
                                        total_shaded_pixels++
                                    endif
                                endif else begin ; middle pixel
                                    if((image[mask_index-1, current_slice] eq 0) and (image[mask_index+1, current_slice] eq 0)) then begin
                                        image [mask_index, current_slice] = 220
                                        image2[mask_index, current_slice] = 1
                                        mask_bit_filled = 1
                                        total_shaded_pixels++
                                    endif
                                endelse
                            endif
                        endif
                    endfor
                endfor
            endfor
        endif

        if(first_shaded_bit_this_slice lt first_shaded_bit) then first_shaded_bit = first_shaded_bit_this_slice
        if( last_shaded_bit_this_slice gt  last_shaded_bit) then  last_shaded_bit =  last_shaded_bit_this_slice

        current_slice++
    endif ; normal uncompress image

    ; L2 and L4 and area accumulators

    if(max(shaded_bits) gt max_shaded_bits) then begin ; for L2, the most shaded bits in a slice
        max_shaded_bits = max(shaded_bits)
    endif

    if(max_slice_width lt last_shaded_bit_this_slice - first_shaded_bit_this_slice + 1) then begin ; for L4
        max_slice_width = last_shaded_bit_this_slice - first_shaded_bit_this_slice + 1
        midpoint = (last_shaded_bit_this_slice - first_shaded_bit_this_slice) / 2. + first_shaded_bit_this_slice ; for PC1, the L2 midpoint, note that this will be the center of the longest section of shaded bits
    endif

    total_shaded_pixels += total(shaded_bits) ; for area

    if(current_slice ge slices) then break
endfor

;if(slices ge 100) then stop

index = where(image[0, *] eq 0, count)
if(count gt 0) then begin
    edges or= '2'X ; image touching first edge flag (pixel 0 shaded)
endif

index = where(image[!PIXELS - 1, *] eq 0, count)
if(count gt 0) then begin
    edges or= '1'X ; image touching last edge flag (pixel 127 shaded)
endif

area = total_shaded_pixels

L2 = max_shaded_bits ; L2, most shaded bits in a slice
L4 = max_slice_width ; L4, the widest single slice

L5 = last_shaded_bit - first_shaded_bit + 1 ; L5, the max width of the particle across all slices


if(area lt 1) then begin ; calculate particle parameters, Roundness, perimeter, etc
    ;print, "no shaded pixels this particle, should reject"


    ;stop
endif else if(area eq 1) then begin ; set defaults for single shaded pixel event
    ; note no need to set the perimeter x, y, or the Length and Width endpoints

    maxLength = 1.0
    maxWidth  = 1.0
    num_perimeter_points = 1L
    perimeterum      = float(microns_per_pixel * 4.) ; TAF check with people on this one, should it be pi or 4?

    Roundness1 = -1.
    Roundness2 = -1.
    Roundness3 = -1.

    num_satellites = 1
    largest_satellite_area = microns_per_pixel * microns_per_pixel

    ; NOTE 2015-04-27 crystal_type and csp were the only values being calculated below for single pixel images, so moving them here and returning
    ; shouldn't increase the speed much if at all, not sure where the bottleneck is, maybe in the call/setup?
    crystal_type = 1   ; sphere
    csp = (largest_satellite_area * maxWidth*microns_per_pixel * (2.*maxLength*microns_per_pixel + 2.*maxWidth*microns_per_pixel)) / perimeterum

    return
endif else begin; at least two shaded pixels
    ; save the edge pixel counts
    edge_index_0   = where(image[  0, *] eq 0, edge_pixel_count_0)
    edge_index_127 = where(image[127, *] eq 0, edge_pixel_count_127)

    ; centroid of on-array image, note this is distinct from the centroid for off-edge circle fit

    Centroid = fltarr(2)

    index = where(image eq 0, count)

    index2xy, !PIXELS, index, x_index, y_index

    Centroid[0] = total(x_index) / Area
    Centroid[1] = total(y_index) / Area

    ; perimeter trace

    ;neighbor            = [-1, -131, -130, -129,   1, 131, 130, 129] ; for linear array
    neighbor = [[-1,0],[-1,1],[0,1],[1,1],[1,0],[1,-1],[0,-1],[-1,-1]] ; for 2D array [2, 8]
    next_neighbor_index = [5, 6, 7, 0, 1, 2, 3, 4]

    OneHalfPI = !pi/2.

    still_looking = 1

    pixel1_index  = [-1, -1]
    pixel2_index  = [-1, -1]
    pixel3_index  = [-1, -1]
    seen_pixel1_again = 0
    angle_ratio = 0.0

    Bimage_width = !PIXELS + 2
    pixel_map_size = (slices+2L) * Bimage_width

    ; Bounded image, the image array, with a one pixel boundary around all sides
    Bimage = bytarr(!PIXELS + 2, slices + 2) ; i.e. width+2 * height+2 to have the blank edge pixels

    code_array = intarr(!PIXELS + 2, slices + 2)

    ; max possible size is all the points. It will likely be less since it is the perimeter
    ; but we are allocating space for maximum # pixels

    Bimage_index   = [-1, -1]
    FirstPixel   = 0L
    neighbor_ptr = 0

    pixel_location = 0

    Bimage[0:!PIXELS-1, 0:slices-1] = image2
    Bimage = shift(Bimage, 1, 1)
    image_array = Bimage

    ; trace around the perimeter
    ; start by finding the topmost, leftmost set pixel. This will be our starting point.

    ; if the perimeter vs area criteria are not met, start with a fresh starting point,
    ; track the longest perimeter, and return it regardless of perimeter criteria being met
    ; set edges |= 0x20 if perimeter vs area criteria are not met
    ; set edges |= 0x40 if there are satellite images (particle image is disjoint)

    ; set the edges flag for perimeter criteria not set,
    ; the flag will be unset below in the perimeter loop iff the criteria are set
    ; TAF 2013-10-10 new edge flags are not yet handled in later processing
    ; edges bits 1 and 2 are compared for detecting non-edge particles,
    ; but they are not masked with edges && 0x03 first, so the detection
    ; gets false positives if the other new edge flags are set, and discards particles
    ;edges or= '20'X

    Bimage_start_index            = 0L
    num_longest_perimeter_points  = 0L
    longest_perimeter_start_index = 0L

    start_slice = 0L

    ; perform image segmentation of the connected components using diagonal connectivity

    simage = label_region(bimage, /all_neighbors)

    ; the maximum of the segmented image is the number of distinct image components (satellites)
    ; note that by this count, a single image is set to 1 for num_satellites
    ; ie Mercury and Venus would be 1, Earth+Moon would be 2, and Mars+Fear+Dread would be 3

    num_satellites = max(simage)

    ; use the histogram function to identify the largest image segment
    ; the reverse_indices will be used to identify a pixel of the largest image segment for the perimeter trace

    h = histogram(simage, reverse_indices = r)

    ; code to remove smaller satellite images
    ;image = label_region(neg(shattered_image1), /all_neighbors)
    ;h = histogram(image, reverse_indices = r)
    ;index = where(h le max_size, count)

    ;for i=0, count-1 do begin ; (protect count=0)
    ;   image[r[r[index[i]]:r[index[i]+1]-1] = 0
    ;endfor


    ; find the largest image segment by finding the area of each satellite image and taking the max
    largest_satellite_area = 0
    largest_satellite_key  = 1 ; initial value = 1, 0 is for the background

    satellite_areas = intarr(num_satellites+1) ; +1 so that satellite_key can be used directly for indexing the areas

    if(num_satellites ge 1) then begin
        ; the intent is that it will always go through the for loop one pass and assign the proper start index
        ; may need to handle an empty image here as well -- though check to see if they are handled before this point
        for satellite_key = 1, num_satellites do begin
            satellite_indices = where(simage eq satellite_key, satellite_area)

            if(satellite_area gt largest_satellite_area) then begin
                largest_satellite_area    = satellite_area
                largest_satellite_key     = satellite_key
                ;largest_satellite_indices = satellite_indices ; faster without this assignment
            endif
        endfor
    endif else begin
        print, 'image with no satellites detected (empty image)'
        return
    endelse

    ; scale the largest_satellite_area to square microns
    largest_satellite_area = float(largest_satellite_area) * microns_per_pixel * microns_per_pixel

    start_x = long(r[r[largest_satellite_key]] mod long(Bimage_width))
    start_y = long(r[r[largest_satellite_key]] /   long(Bimage_width))

    pixel1_index  = [-1, -1]
    pixel2_index  = [-1, -1]
    pixel3_index  = [-1, -1]
    seen_pixel1_again = 0
    num_perimeter_points = 0L
        perimeter_points = dblarr(2, area)

    Bimage_index[0] = start_x
    Bimage_index[1] = start_y

    FirstPixel         = Bimage_index
    pixel1_index       = Bimage_index
    Bimage_start_index = Bimage_index

    perimeter_points[0, num_perimeter_points] = double(Bimage_index[0] - 1)
    perimeter_points[1, num_perimeter_points] = double(Bimage_index[1] - 1)
    num_perimeter_points = 1L ; or num_perimeter_points++? though this is the first point...

    Bimage[Bimage_index[0], Bimage_index[1]]++ ; flag that we've seen this point as our starting position
    neighbor_ptr = 0 ; start neighbor search in position #1 (i.e. neighbor[0])

    ;  to make the perimeter routine avoid some odd configurations that result in missing
    ;  some of the perimeter points, you need to continue to cycle around the perimeter
    ;  until you have seen the "starting point" and then the 2nd point a second time.
    ;  Below is an example of a particle
    ;  where you would see the "starting point" (indicated by the S) two times in this algorithm
    ;  and would not have accounted for all the points. Thus you need to see the 2nd pixel again
    ;  to be sure you are really done.

    ;      SX
    ;    XX
    ;    XX

    ; TAF note, the above explanation is insufficient and the code was still missing pixels

    still_looking = 1
    while(still_looking) do begin ; Main perimeter tracking loop
        ; cycle around the neighbor locations looking for the next potential perimeter pixel

        found_neighbor = 0

        ; might IDLize the for below with something like:
        ; index = where(Bimage[Bimage_index[0] + neighbor[0, neighbor_ptr], $
        ;               Bimage_index[1] + neighbor[1, neighbor_ptr]] eq 0, count)
        ; if count gt 0 then begin
        ;    Bimage_index[0] += neighbor[0, neighbor_ptr??? how does this work again??

        for i=0, 7 do begin ; set up next neighbor
            ; If no pixel in this direction, try again in next clockwise position
            if(Bimage[Bimage_index[0] + neighbor[0, neighbor_ptr], $
                      Bimage_index[1] + neighbor[1, neighbor_ptr]] eq 0) then begin
                neighbor_ptr = (neighbor_ptr + 1) mod 8
            endif else begin
                found_neighbor = 1 ; found a non-zero neighbor
                ;break??? or do we let neighbor_ptr wrap the whole way?
            endelse
        endfor

        if(found_neighbor) then begin ; meaning we found a neighbor pixel that is set
            ; set Bimage_ptr to new pixel position
            Bimage_index[0] += neighbor[0, neighbor_ptr]
            Bimage_index[1] += neighbor[1, neighbor_ptr]

            if(Bimage_index[0] eq pixel1_index[0] and Bimage_index[1] eq pixel1_index[1]) then begin
                seen_pixel1_again = 1
            endif

            ; are we back at the location of pixel #2, after having stepped on #1 again?
            if(Bimage_index[0] eq pixel2_index[0] and Bimage_index[1] eq pixel2_index[1] and seen_pixel1_again eq 1) then begin
                still_looking = 0
            endif else begin
                if(num_perimeter_points eq area) then begin ; TAF horizontal lines kept tracing around, crashing IDL from the dll implementation
                    still_looking = 0
                endif else begin
                    if(Bimage[Bimage_index[0], Bimage_index[1]] eq 1) then begin
                        ; we have an unflagged pixel for the perimeter, add it to our list
                        perimeter_points[0, num_perimeter_points] = double(Bimage_index[0] - 1)
                        perimeter_points[1, num_perimeter_points] = double(Bimage_index[1] - 1)
                        num_perimeter_points++

                        if(num_perimeter_points gt area) then begin
                            ;return -7; // TAF num_perimeter_points >= NumPoints
                            print, "num_perimeter_points > area"
                            stop
                        endif
                    endif

                    Bimage[Bimage_index[0], Bimage_index[1]]++ ; flag that this perimeter point as being seen (possibly again)

                    neighbor_ptr = next_neighbor_index[neighbor_ptr];  // select our next starting location for the neighbor search

                    if((num_perimeter_points eq 2L) and (pixel2_index[0] eq -1L)) then begin
                        pixel2_index = Bimage_index ; when we hit the 2nd point save its index. when we see it again, we're done
                    endif
                endelse
            endelse
        endif else begin ; we didn't find a neighbor
            still_looking = 0 ; then we're done
        endelse
    endwhile ; Main perimeter tracking loop

    perimeter_points_x = reform(perimeter_points[0,0:num_perimeter_points-1]) ; This step could be skipped by changing the proceedure arguments to match the 2D array
    perimeter_points_y = reform(perimeter_points[1,0:num_perimeter_points-1])

    ; have the perimeter traced, now calculate the length and width

    maxLength = -1.
    maxWidth  = -2.
    angle_tolerance = .075 ; radians, 4.3 degrees
    size_adjust = 5.5 ; pixels below this length, we ease the angle_tolerance by a factor of 2
                      ; this works reasonably well, but should do an analysis on actual angle distributions with size/geometery

    if(num_perimeter_points ge 2L) then begin
        ; get max length chord
        ; get max  width chord

        distance_chords = abs(distance_measure(perimeter_points[*,0:num_perimeter_points-1L], /matrix)) + 1.

        max_distance_index = where(distance_chords eq max(distance_chords), count)

        if(count gt 0) then begin

            index2xy, num_perimeter_points, max_distance_index[0], ml_p1, ml_p2

            ; the x and y are then the indices into perimeter_points of the two points with the longest distance

            maxLength = distance_chords[max_distance_index[0]]

            ml_p1x = perimeter_points[0,ml_p1]
            ml_p1y = perimeter_points[1,ml_p1]

            ml_p2x = perimeter_points[0,ml_p2]
            ml_p2y = perimeter_points[1,ml_p2]

            if(num_perimeter_points eq 2L) then begin ; just assign a 1 for the width
                maxWidth = 1.

                mw_p1x = perimeter_points[0,0]
                mw_p1y = perimeter_points[1,0]
                mw_p2x = perimeter_points[0,0]
                mw_p2y = perimeter_points[1,0]
            endif else begin ; we have 3 or more perimeter points, find the width chord
                ; find the widest chord of the event in pixels, at 90 degrees (+-.075 radians, 4.3 degrees) to the longest chord (maxLength)
                ; first find the angle of the max Length line from the two endpoints, ml_p1, and ml_p2

                if(ml_p1 lt ml_p2) then begin
                    first  = ml_p1
                    second = ml_p2
                endif else if(ml_p1 gt ml_p2) then begin
                    first  = ml_p2
                    second = ml_p1
                endif else begin ; points the same, assume 1 pixel width, but should never get here
                    print, "width chord finding error, points the same
                    stop
                endelse

                p1x = perimeter_points[0,first]
                p1y = perimeter_points[1,first]
                p2x = perimeter_points[0,second]
                p2y = perimeter_points[1,second]

                dx = p2x - p1x
                dy = p2y - p1y

                angle = atan(dy, dx) ; IDL atan returns OneHalfPI for dx == 0

                ; TAF should be an easier way, add 2.5 pi, mod pi -- but no mod for doubles in c++
                ; mod pi seems to work in IDL, but need to change algorithm for negative numbers

                angle = fix(modpi(angle + OneHalfPI) * 1000.)/1000. ; rotate by 90 degrees, constrain sig figs to avoid floating errors

                ; search along chords with this orientation for the largest distance at a right angle to Length (ie. the width)

                ; calculate the angles from each perimeter point on one side of the Length chord to all other perimeter points
                ; on the other side of the Length chord
                ; then find the max length chord within some tolerance (.1) to perpendicular to the maxLength chord

                ; NOTE: the perimeter is not necessarily well ordered, so tracing either side of the Length is not sufficient to search

                ; the Width for loop below is based on the CPI View Width

                ; default to the first pixel and with width = 1
                maxWidth = 1.
                mw_p1x = perimeter_points[0,0]
                mw_p1y = perimeter_points[1,0]
                mw_p2x = perimeter_points[0,0]
                mw_p2y = perimeter_points[1,0]

                ; adjust the angle tolerance for small particles
                if(L5 lt size_adjust or slices lt size_adjust) then begin
                    angle_tolerance = .15
                endif else begin
                    angle_tolerance = .075
                endelse

                for i=0L, num_perimeter_points-2 do begin
                    dx = long(perimeter_points[0, i+1:num_perimeter_points-1] - perimeter_points[0, i])
                    dy = long(perimeter_points[1, i+1:num_perimeter_points-1] - perimeter_points[1, i])

                    angle_now = atan(dy, dx)

                    index = where(angle_now lt 0, count)
                    if(count gt 0) then angle_now[index] += !PI

                    index = where(angle_now gt !PI, count)
                    if(count gt 0) then stop ;angle_now[index] -= !PI

                    ;index = where(abs(angle_now - angle) lt angle_tolerance, count) ; did not take all cases into account, and could suffer subtraction errors

                    ; get rid of small errors in floating point calculations
                    mod_angle     = fix(modpi(angle     + angle_tolerance) * 1000.)/1000.
                    mod_angle_now = fix(modpi(angle_now + angle_tolerance) * 1000.)/1000.

                    index = where(((angle_now lt angle) and (angle_now + angle_tolerance ge angle    ))   or  $  ; angle - tol > angle_now
                                  ((angle lt angle_now) and (angle     + angle_tolerance ge angle_now))   or  $  ; angle + tol < angle_now
                                  ((angle_now le angle_tolerance) and (mod_angle     le angle_tolerance)) or  $  ; angle_now close to   0, angle close to 180
                                  ((angle     le angle_tolerance) and (mod_angle_now le angle_tolerance)) or  $  ; angle_now close to 180, angle close to   0
                                  ((angle le angle_tolerance)     and (angle_now     le angle_tolerance)) or  $  ; angle_now close to   0, angle close to   0
                                  ((mod_angle le angle_tolerance) and (mod_angle_now le angle_tolerance)), count); angle_now close to 180, angle close to 180

                    if(count gt 0) then begin
                        len = distance_chords[i, i + 1 + index]

                        maxWidth = max(len) > maxWidth

                        subindex = where(len eq maxWidth, count)

                        if(count gt 0) then begin
                            mw_p1x = perimeter_points[0,i]
                            mw_p1y = perimeter_points[1,i]
                            mw_p2x = perimeter_points[0,i+1+index[subindex[0]]]
                            mw_p2y = perimeter_points[1,i+1+index[subindex[0]]]
                        endif
                    endif
                endfor
            endelse
        endif

        ; calculate perimeter length in microns using lookup table

        ; Calculate the length of the perimeter in microns, using the lookup table method

        perimeter_length = 0.

        image_index = 0L

        ; certainly a way to make this more IDL like to speed it up

        for ii=0L, num_perimeter_points-1L do begin
            image_index_x = long(perimeter_points[0,ii])
            image_index_y = long(perimeter_points[1,ii])

            for j=0, 7 do begin
                search_index_x = image_index_x + neighbor[0,j]
                search_index_y = image_index_y + neighbor[1,j]

                ; TAF these checks might be excessively conservative here
                if(search_index_x gt 0 and search_index_x lt !PIXELS+2 and $
                   search_index_y gt 0 and search_index_y lt slices+2) then begin
                    if(image_array[search_index_x, search_index_y] eq 1) then begin
                        code_array[image_index_x, image_index_y] += neighbor_codes[j]
                    endif
                endif
            endfor

            if(code_array[image_index] ge 256) then begin
                ;return -8; // TAF code_array[image_index] >= 256
                print, "perimeter length calculation error code_array[image_index] >= 256"
                break
            endif

            perimeter_length += length_table[code_array[image_index_x, image_index_y]]
        endfor

        perimeterum = float(microns_per_pixel * perimeter_length); // convert for 10 microns per pixel

    endif else begin ; TAF should actually never get in here as single pixel events are handled above?
        ; copy over the Length and Width end points for a single pixel event
        ml_p1x = perimeter_points[0,0]
        ml_p1y = perimeter_points[1,0]

        ml_p2x = perimeter_points[0,0]
        ml_p2y = perimeter_points[1,0]

        mw_p1x = perimeter_points[0,0]
        mw_p1y = perimeter_points[1,0]

        mw_p2x = perimeter_points[0,0]
        mw_p2y = perimeter_points[1,0]

        maxLength = 1.
        maxWidth  = 1.
        perimeterum = float(microns_per_pixel * 4.0);  mo added this , it does come to this point.

    endelse

    ; Roundness1
    ; Roundness2
    ; Roundness3

    ; ========= Roundness1 ================
    if(maxLength ne -1.) then begin
        Roundness1 = (4.*Area)/(!PI*maxLength^2.)
    endif else begin
        Roundness1 = -1.
    endelse

    if(num_perimeter_points gt 2L) then begin
        ;radii = fltarr(num_perimeter_points)

        ;for i=0, num_perimeter_points-1 do begin
        ;    radii[i] = sqrt((float(Centroid[0]) - float(BoundaryPtsX[i]))^2. + (float(Centroid[1]) - float(BoundaryPtsY[i]))^2.)
        ;endfor

        radii = sqrt((float(Centroid[0]) - float(reform(perimeter_points[0,0:num_perimeter_points-1])))^2. + $
                     (float(Centroid[1]) - float(reform(perimeter_points[1,0:num_perimeter_points-1])))^2.)
    endif

    Roundness2 = -1.
    Roundness3 = -1.

    if(num_perimeter_points gt 4L) then begin
        estimated_radius = 0.5 * maxLength

        dist_variance = 0.

        ;for i=0, num_perimeter_points-1 do begin ; TAF should be able to IDLize this for loop, transfer/merge with CPI
        ;    diff = radii[i] - estimated_radius
        ;    dist_variance = dist_variance + diff^2.
        ;endfor

        diff = radii - estimated_radius
        dist_variance = total(diff^2.) / (num_perimeter_points-1)

        Roundness2 = sqrt(dist_variance) / estimated_radius ; convert to SD and scale to the presumed radius

        ; ========= Roundness3 ================
        ; Algorithm based on Chris Wieland's June 2012 Roundness 3 implemented for the 2D-S Process Particles dll

        ; calculates the ratio of the average angles of the perimeter pixels with respect to the centroid, and compares
        ; to the expected angles for a circle

        one = 0
        two = 1

        ; TAF catch the NaN on larger particles, div by zero???

        angles = fltarr(num_perimeter_points)

        estimated_angle  = 360. / num_perimeter_points
        angle_variance   = 0.
        mean_angle       = 0.

        for i=0, num_perimeter_points-1 do begin ; TAF might be able to IDLize this loop as well, could be dependent on getangle?
            angles[i] = getangle(perimeter_points[0,one], perimeter_points[1,one], Centroid[0], Centroid[1], perimeter_points[0,two], perimeter_points[1,two])

            one = (one+1) mod num_perimeter_points
            two = (two+1) mod num_perimeter_points

            mean_angle = mean_angle + angles[i]
        endfor

        mean_angle /= num_perimeter_points

        diff = angles - estimated_angle
        angle_variance = total(diff^2.) / (num_perimeter_points-1)

        Roundness3 = sqrt(angle_variance) / estimated_angle ; convert to SD and scale to the presumed angle
    endif ; (num_perimeter_points gt 4)
endelse ; NumPoints in image >= 2

; if round, or single pixel, set crystal type to 1, sphere, else set to 15, ice
if((Roundness1 gt !Roundness1_min and $
    Roundness1 lt !Roundness1_max and $
    Roundness2 gt !Roundness2_min and $
    Roundness2 lt !Roundness2_max and $
    Roundness3 gt !Roundness3_min and $
    Roundness3 lt !Roundness3_max) or $
    (Area eq 1)) then begin
    crystal_type = 1   ; sphere
endif else begin
    crystal_type = 15  ; ice
endelse

; calculate the Combined Single Parameter CSP in um^3
; Area, maxWidth, and maxLength are in pixels
; perimeterum is in microns

if(perimeterum eq 0.) then begin
    csp = -1.
endif else begin
    csp = (largest_satellite_area * maxWidth*microns_per_pixel * (2.*maxLength*microns_per_pixel + 2.*maxWidth*microns_per_pixel)) / perimeterum
endelse

; generate the recompressed image with mask bits filled for the SORTED file, put the adjusted wordcount into new_wordcount
if(!mask_bit_filling and mask_bit_filled ne 0) then begin

    ; check (noncompressed_slice_count gt 0) here???
    recom_words = uintarr(8L * wordcount) ; for building the recompressed image data

    current_word = 0L

    for i=0L, slices-1L do begin
        recom_words[current_word] = recom_words[current_word] or '4000'X ; label start of slice

        shade_count = 0U
        clear_count = 0U

        counting_shade = 0

        ; scan through the slice, counting contiguous shaded or clear sections
        ; put the counts into the recompressed image, clear bits then shaded bits
        for j=0, !PIXELS-1 do begin
            if(image[j, i] eq 0 or image[j, i] eq 220) then begin ;
                counting_shade = 1

                shade_count++

                if(shade_count eq !PIXELS) then begin ; full shaded slice -- leave as 0x4000
                    break
                endif

                if(j eq !PIXELS-1) then begin ; shaded pixels against high side of image, write out word
                    if(current_word ge (size(recom_words))[1]) then begin
                        print, 'recompress failure in shaded against high side, current_word = ', current_word, ' / ', (size(recom_words))[1]
                        break
                    endif
                    recom_words[current_word] = recom_words[current_word] or ishft(shade_count, 7) or clear_count
                    current_word++
                endif
            endif else if(image[j, i] ne 0 and image[j, i] ne 220) then begin
                if(counting_shade) then begin ; transitioned from shade to clear, generate a word
                    counting_shade = 0
                    if(current_word ge (size(recom_words))[1]) then begin
                        print, 'recompress failure in shade to clear, current_word = ', current_word, ' / ', (size(recom_words))[1]
                        break
                    endif
                    recom_words[current_word] = recom_words[current_word] or ishft(shade_count, 7) or clear_count
                    current_word++
                    clear_count = 0
                    shade_count = 0
                endif

                clear_count++
            endif
        endfor

        if(current_word gt (size(recom_words))[1]) then begin
            print, 'recompress failure exiting loop, current_word = ', current_word, ' / ', (size(recom_words))[1]
            stop
            break
        endif
    endfor

;if(slices ge 100) then stop

    ; cast the words into a bytearr

    if(current_word eq 0) then begin
        print, 'recompress failure no data added, current_word = ', current_word, ' / ', (size(recom_words))[1]
        current_word = 1
    endif else if(current_word ge (size(recom_words))[1]) then begin
        print, 'recompress failure exceeded recom_words, current_word = ', current_word, ' / ', (size(recom_words))[1]
        current_word = (size(recom_words))[1]-1
    endif

    new_wordcount = current_word

    index = indgen(new_wordcount)

    hi_bytes = byte(ishft(recom_words[index], -8) and '00FF'X)
    lo_bytes = byte(      recom_words[index]      and '00FF'X)

    recompressed_image = bytarr(2L * new_wordcount)

    recompressed_image[2*index+1] = hi_bytes
    recompressed_image[2*index  ] = lo_bytes

    ; put the image and new wordcount into the calling arguments

    image_data = recompressed_image
    ;image already using the export variable
    WordCount = new_wordcount
endif


;stop

END ; ProcessParticle
