; File: read_hk_3v.pro
; Author:  Ted Fisher <ted@specinc.com> <fishnuke@gmail.com>
; Purpose:  Reads a 3V-CPI type HK file and generates plots
;           for 3V type HK as generated by the 3V-2DS or uCOPP


; version 1.4 2016-02-04 added output csv option, with probe-specific HK outputs    NOTE: need to test on uCOPP data
; version 1.3 2016-01-27 added mask_count to parameter list
; version 1.2 2015-12-14 modified for white background plots
; version 1.1 2015-12-08 changed file and main procedure to _3v as opposed to _2d for 2D-S type HK
; version 1.0 2014-04-28 initial version




;fix the HK plot second axes for white background

;change main PSD not to use window 2 (used for HK)


function counts2temp_3V, in_array
    ; convert raw data counts to temperature using the following conversion

    ; rt = 6.5536e9 * (1.-double(in_array/65536.) / (5. * double(in_array))
    ; temp_fst = -273.15 + (1.1117024e-3 + 237.02702e-6 * alog(rt) + 75.78814e-9 * (alog(rt))^3.)^(-1.)

    ; translated from the matlab HK read code in t2vcpi_hk.m

    ; rt = 6.5536e9 *  (1-data(HKDataIDX+1)/65536)/(5*data(HKDataIDX+1))
    ; temp_fst = [temp_fst, -273.15 + (1.1117024e-3 + 237.02702e-6*log(rt)+75.78814e-9*(log(rt))^3)^(-1)]; %Forward Sample Tube

    rt = 6.5536e9 * (1.-double(in_array)/65536.) / (5. * double(in_array))
    temp_array = -273.15 + (1.1117024e-3 + 237.02702e-6 * alog(rt) + 75.78814e-9 * (alog(rt))^3.)^(-1.)

    return, temp_array
end ; counts2temp_3V



pro read_hk_3v, basefile=basefile, H50micron=H50micron, hk_count=hk_count, mask_count=mask_count, show_plots=show_plots, save_plots=save_plots
    ; read a 3V-CPI Housekeeping file into memory, generate arrays of HK, Masks, and timestamps, and convert the raw
    ; HK counts into physical values

    ; fills the basescan_common hk_ arrays

    ; hk_count is the hk_ array size

    ; H50micron is set to 1 for H 50 micron probes, 0 for H 10 micron probes

    ; setting /show_plots to 1 shows plots of HK data, setting to 0 overrides /save_plots
    ; setting /save_plots saves the HK plots

    ; note, should update to check second word = 83 and the checksums, and make parsing more robust

    ; need to make sure hk_array and hk_timestamp_array match up -- maybe put the timestamp struct in the hk_array???

    ; need to detect H 50 micron from Housekeeping and put it into output files/sets flag H50micron

    ;    (from the c++)
    ;    // test for the Hawkeye 50 micron flag in HK Word 81, bit 15
    ;    if((hk->data2[3] & 0x8000) != 0)
    ;    {
    ;        *H50micron_flag = 1
    ;        H50micron = true
    ;    }
    ;    else
    ;    {
    ;        *H50micron_flag = 0
    ;        H50micron = false
    ;    } // end if



    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp, buffer, TAS_array

    ;setup_globals ; disable this call when read_hk is incorporated into basescan

    if(keyword_set(show_plots)) then begin
        show_hk_plots = show_plots
    endif else begin
        show_hk_plots = 1
    endelse

    if(keyword_set(save_plots)) then begin
        save_hk_plots = save_plots
    endif else begin
        save_hk_plots = 0
    endelse

    hk_plotsize_x = 700
    hk_plotsize_y = 400

    plot_against_time = 1


    ;    hpos = 512
    ;    vpos = 512

    ;    reprint_raw_data = 1

    ;    inword = uint(0) ; read 2 bytes
    ;    Hwords = uint(0)
    ;    Vwords = uint(0)

    ;    total_Hpart     = 0L
    ;    total_Vpart     = 0L
    ;    total_Hmulti    = 0L
    ;    total_Vmulti    = 0L
    ;    total_Hoverflow = 0L
    ;    total_Voverflow = 0L
    ;    total_Hcpi_trigger = 0L
    ;    total_Vcpi_trigger = 0L
    ;    total_Hflag_bit13  = 0L
    ;    total_Vflag_bit13  = 0L
    ;    total_Hflag_bit12  = 0L
    ;    total_vflag_bit12  = 0L
    ;    total_NH_and_NV_ne_zero = 0L
    ;    total_NULL_count   = 0L

    ;        Hpart_this_second = 0L
    ;        Vpart_this_second = 0L
    ;       Hmulti_this_second = 0L
    ;       Vmulti_this_second = 0L
    ;    Hoverflow_this_second = 0L
    ;    Voverflow_this_second = 0L
    ;         NULL_this_second = 0L

    ;    probecountflag_set = 0 ; debug flag for stopping at specified probecount values

    ;    probecount = long64(0)


    if(n_elements(basefile) eq 0) then begin
        ; hard coded input for faster testing, need the full file and path names
        ;infilename = 'C:\fisher\SPEC\3VCPI\basescan\test_data\base130909131733.2DSCPIHK'
        ;infilepath = 'C:\fisher\SPEC\3VCPI\basescan\test_data\'

        ; Open a dialog window to select the base.2DSCPI file to open, and retrieve the path
        infilename = dialog_pickfile(title='Select 3V-CPI 2D-S Base HK File', filter='base*.2DSCPIHK', get_path=infilepath)

        if(infilename eq '') then begin ; user cancelled open dialog
            print, ''
            print, 'User Cancelled opening 3V-CPI type base*.2DSCPIHK file, read_hk_3v exiting.'
            print, ''

            return
        endif
    endif else begin
        infilename = strcompress(basefile + 'HK', /rem)
        infilepath = strmid(basefile, 0, strpos(basefile, '\', /reverse_search) + 1)
    endelse

    stripped_infilename = strmid(infilename, strlen(infilepath)) ; the file name with the path removed, for generating the output file names


    ;    print, ''
    ;    print, 'read_hk opening file: ', stripped_infilename
    ;    print, ''
    ;    print, 'Scan starting at:'
    ;    print, systime()

    startseconds = systime(/seconds)



    openr, lun, /get, infilename

    infilefstat = fstat(lun)

    hk_file_size = infilefstat.size

    hk_lun = lun

    ;print,hk_lun
    ;stop


    ; Housekeeping ASCII Flag 'HK' (0x484B)
    ; Mask         ASCII Flag 'MK' (0x4D4B)




    flag = uint(0)

    hk_words = 82 ; 83 including ID, second word is Packet Length, which should = 83
    mask_words = 27

    hk_frame = uintarr(hk_words)
    mask_frame = uintarr(mask_words)


    array_size = !array_growth_factor

    empty_timestamp = timestamp

    hk_timestamp_array = replicate(empty_timestamp, array_size)

    hk_array = uintarr(array_size,   hk_words)
    mask_array = uintarr(array_size, mask_words)

    hk_count = 0L
    mask_count = 0L
    timestamp_count = 0L


    done = 0

    while(not done) do begin
        ; read the 16 byte MS timestamp
        ; test explicitly for end of file before each read
        point_lun, -hk_lun, current_pos

        if(current_pos eq hk_file_size) then begin
            print, 'at end of file, exiting'
            break
        endif

        next_read_length = n_tags(timestamp, /data_length)
        if(current_pos + next_read_length gt hk_file_size) then begin
            print, 'at end of file, exiting'
            break
        endif

        readu, hk_lun, timestamp

        hk_timestamp_array[timestamp_count] = timestamp

        timestamp_count++

        point_lun, -hk_lun, current_pos

        next_read_length = n_tags(timestamp, /data_length)
        if(current_pos + next_read_length gt hk_file_size) then begin
            print, 'at end of file, exiting'
            break
        endif

        point_lun, -hk_lun, current_pos

        next_read_length = n_tags(flag, /data_length)
        if(current_pos + next_read_length gt hk_file_size) then begin
            print, 'at end of file, exiting'
            break
        endif

        readu, hk_lun, flag

        if(flag eq '484B'X) then begin ; 0x484B HK Housekeeping
            ;print, 'HK found'

            point_lun, -hk_lun, current_pos

            next_read_length = n_tags(hk_frame, /data_length)
            if(current_pos + next_read_length gt hk_file_size) then begin
                print, 'at end of file, exiting'
                break
            endif

            readu, hk_lun, hk_frame

            hk_array[hk_count, *] = hk_frame

            hk_count++
        endif else if(flag eq '4D4B'X) then begin ; 0x4D4B MK Mask
            ;print, 'Mask found ', mask_count+1

            point_lun, -hk_lun, current_pos

            next_read_length = n_tags(mask_frame, /data_length)
            if(current_pos + next_read_length gt hk_file_size) then begin
                print, 'at end of file, exiting'
                break
            endif

            readu, hk_lun, mask_frame

            mask_array[mask_count, *] = mask_frame

            mask_timestamp_array[mask_count] = timestamp

            mask_count++


            ; Word       Mask Data Description
            ; 1          Flag ASCII "MK" (0x4D4B)
            ; 2          Packet Length (in words) - 28 (0x1C)
            ; 3-5        48 bit Timing Word (read from the FPGA)
            ; 6-13       Horizontal mask bits -  Directly Read from FPGA (hz_mask_?)
            ; 14 - 21    Vertical mask bits - Directly Read from FPGA (vt_mask_?)
            ; 22 - 24    Timing Word at start of mask routine - Directly Read from FPGA (tw_?sb)
            ; 25 - 27    Timing Word at end of mask routine - Directly Read from FPGA (tw_?sb)
            ; 28         Checksum Word


        endif

        ; test the counts against the array size and expand all arrays if needed
        if(timestamp_count ge array_size-1 or hk_count ge array_size-1 or mask_count ge array_size-1) then begin
            ;stop

            array_size += 10000L

            if(timestamp_count gt 0) then begin
                new_timestamp_array = replicate(empty_timestamp, array_size)
                new_timestamp_array[0:timestamp_count-1] = temporary(hk_timestamp_array[0:timestamp_count-1])
                hk_timestamp_array = temporary(new_timestamp_array)
            endif

            if(hk_count gt 0) then begin
                new_hk_array = uintarr(array_size, hk_words)
                new_hk_array[0:hk_count-1, *] = temporary(hk_array[0:hk_count-1, *])
                hk_array = temporary(new_hk_array)
            endif

            if(mask_count gt 0) then begin
                new_mask_array = uintarr(array_size, mask_words)
                new_mask_array[0:mask_count-1, *] = temporary(mask_array[0:mask_count-1, *])
                mask_array = temporary(new_mask_array)

                new_mask_timestamp_array = replicate(empty_timestamp, array_size)
                new_mask_timestamp_array[0:mask_count-1, *] = temporary(mask_timestamp_array[0:mask_count-1, *])
                mask_timestamp_array = temporary(new_mask_timestamp_array)
            endif

            ;stop
        endif
    endwhile

    free_lun, hk_lun

    ; truncate the arrays to the actual counts
    if(timestamp_count gt 0) then begin
        new_timestamp_array = replicate(empty_timestamp, timestamp_count)
        new_timestamp_array[0:timestamp_count-1] = temporary(hk_timestamp_array[0:timestamp_count-1])
        hk_timestamp_array = temporary(new_timestamp_array)
    endif

    if(hk_count gt 0) then begin
        new_hk_array = uintarr(hk_count, hk_words)
        new_hk_array[0:hk_count-1, *] = temporary(hk_array[0:hk_count-1, *])
        hk_array = temporary(new_hk_array)
    endif

    if(mask_count gt 0) then begin
        new_mask_array = uintarr(mask_count, mask_words)
        new_mask_array[0:mask_count-1, *] = temporary(mask_array[0:mask_count-1, *])
        mask_array = temporary(new_mask_array)

        new_mask_timestamp_array = replicate(empty_timestamp, mask_count)
        new_mask_timestamp_array[0:mask_count-1, *] = temporary(mask_timestamp_array[0:mask_count-1, *])
        mask_timestamp_array = temporary(new_mask_timestamp_array)
    endif

    print, ''
    print, 'Read ',        hk_count, ' HK packets'
    print, 'Read ',      mask_count, ' mask packets'
    print, 'Read ', timestamp_count, ' timestamps'
    print, ''
    ;stop

    ; convert the raw counts to physical units

    temp_fst   = counts2temp_3V(hk_array[*,  1]) ; Forward Sample Tube Temp
    temp_uob   = counts2temp_3V(hk_array[*,  2]) ; Upper Optics Block temp
    temp_lob   = counts2temp_3V(hk_array[*,  3]) ; Lower Optics Block temp
    temp_cst   = counts2temp_3V(hk_array[*,  4]) ; Central Sample Tube temp
    temp_ast   = counts2temp_3V(hk_array[*,  5]) ; Aft Sample Tube temp
    temp_pyln1 = counts2temp_3V(hk_array[*,  6]) ; Pylon 1 temp
    temp_pyln2 = counts2temp_3V(hk_array[*,  7]) ; Pylon 2 temp
    temp_pyln3 = counts2temp_3V(hk_array[*,  8]) ; Pylon 3 temp
    temp_ccd   = counts2temp_3V(hk_array[*,  9]) ; CCD Camera temp
    temp_il    = counts2temp_3V(hk_array[*, 10]) ; Imaging Lens Temp
    temp_ilz   = counts2temp_3V(hk_array[*, 11]) ; Imaging Laser Temp

    ; apparently the lasers use a different temperature sensor -- not sure where this info came from, but doesn't seem accurate all the time...
    ;temp_vlz = double(hk_array[*, 12]) * 0.010173 - 273.15 ; Vertical Laser Temp
    ;temp_hlz = double(hk_array[*, 13]) * 0.010173 - 273.15 ; Horizontal Laser temp

    temp_vlz   = counts2temp_3V(hk_array[*, 12])
    temp_hlz   = counts2temp_3V(hk_array[*, 13])

    temp_pbd   = counts2temp_3V(hk_array[*, 14]) ; Power Board temp
    temp_vpl   = counts2temp_3V(hk_array[*, 15])
    temp_vop   = counts2temp_3V(hk_array[*, 16])
    temp_hpl   = counts2temp_3V(hk_array[*, 17])
    temp_hop   = counts2temp_3V(hk_array[*, 18])
    temp_vmr   = counts2temp_3V(hk_array[*, 19])
    temp_hmr   = counts2temp_3V(hk_array[*, 20])
    temp_ipl   = counts2temp_3V(hk_array[*, 21])
    temp_dsp   = counts2temp_3V(hk_array[*, 22])
    temp_vayt  = counts2temp_3V(hk_array[*, 23])
    temp_vayb  = counts2temp_3V(hk_array[*, 24])
    temp_hayt  = counts2temp_3V(hk_array[*, 25])
    temp_hayb  = counts2temp_3V(hk_array[*, 26])

    rh         = double(hk_array[*, 27]) * 0.002515 - 0.2802198  ; Relative Humidity
    can_pres   = double(hk_array[*, 28]) * 0.000572205 -3.75     ; Internal pressure

    vrt_tec_i  = double(hk_array[*, 29]) * 0.000050498
    hrz_tec_i  = double(hk_array[*, 30]) * 0.000050498
    vrt_lsr    = double(hk_array[*, 31]) * 0.000076294
    hrz_lsr    = double(hk_array[*, 32]) * 0.000076294
    raw_plusv  = double(hk_array[*, 33]) * 0.000152588
    raw_minusv = 2. * double(hk_array[*, 33]) * 0.000152588 - double(hk_array[*, 34]) * 0.00022889

    ;    v_ve0   = double(hk_array[*, 35]) * 0.00244140625
    ;    v_ve21  = double(hk_array[*, 36]) * 0.00244140625 ; Vertical element  21 voltage
    ;    v_ve42  = double(hk_array[*, 37]) * 0.00244140625 ; Vertical element  42 voltage
    ;    v_ve64  = double(hk_array[*, 38]) * 0.00244140625
    ;    v_ve85  = double(hk_array[*, 39]) * 0.00244140625 ; Vertical element  85 voltage
    ;    v_ve106 = double(hk_array[*, 40]) * 0.00244140625 ; Vertical element 106 voltage
    ;    v_ve127 = double(hk_array[*, 41]) * 0.00244140625

    ;    img_lsr_i = double(hk_array[*, 42]) * 0.0268555

    ;    v_he0   = double(hk_array[*, 43]) * 0.00244140625
    ;    v_he21  = double(hk_array[*, 44]) * 0.00244140625 ; Horizontal element  21 voltage
    ;    v_he42  = double(hk_array[*, 45]) * 0.00244140625 ; Horizontal element  42 voltage
    ;    v_he64  = double(hk_array[*, 46]) * 0.00244140625
    ;    v_he85  = double(hk_array[*, 47]) * 0.00244140625 ; Horizontal element  85 voltage
    ;    v_he106 = double(hk_array[*, 48]) * 0.00244140625 ; Horizontal element 106 voltage
    ;    v_he127 = double(hk_array[*, 49]) * 0.00244140625


    ; TAF 2014-05-19 the Horizontal and Vertical element voltages were reversed in the documentation
    ; vs the actual HK from the probe, so they are switched accordingly below
    v_he0   = double(hk_array[*, 35]) * 0.00244140625
    v_he21  = double(hk_array[*, 36]) * 0.00244140625 ; Horizontal element  21 voltage
    v_he42  = double(hk_array[*, 37]) * 0.00244140625 ; Horizontal element  42 voltage
    v_he64  = double(hk_array[*, 38]) * 0.00244140625
    v_he85  = double(hk_array[*, 39]) * 0.00244140625 ; Horizontal element  85 voltage
    v_he106 = double(hk_array[*, 40]) * 0.00244140625 ; Horizontal element 106 voltage
    v_he127 = double(hk_array[*, 41]) * 0.00244140625

    img_lsr_i = double(hk_array[*, 42]) * 0.0268555

    v_ve0   = double(hk_array[*, 43]) * 0.00244140625
    v_ve21  = double(hk_array[*, 44]) * 0.00244140625 ; Vertical element  21 voltage
    v_ve42  = double(hk_array[*, 45]) * 0.00244140625 ; Vertical element  42 voltage
    v_ve64  = double(hk_array[*, 46]) * 0.00244140625
    v_ve85  = double(hk_array[*, 47]) * 0.00244140625 ; Vertical element  85 voltage
    v_ve106 = double(hk_array[*, 48]) * 0.00244140625 ; Vertical element 106 voltage
    v_ve127 = double(hk_array[*, 49]) * 0.00244140625

    img_lsr_pw      = double(hk_array[*, 50]) * 0.0268555
    img_lsr_i_set   = double(hk_array[*, 51]) * 0.014648
    img_lsr_pw_set  = double(hk_array[*, 52]) * 0.014648
    probe_mode_raw  =        hk_array[*, 53]  ; bits 6-7 CPI Trigger Source
    heater_status   = double(hk_array[*, 54]) * 1.
    optical_pwm_pct = double(hk_array[*, 55]) * 0.0268555
    num_hrz_dect    =        hk_array[*, 56]  ; Horizontal particles detected
    num_vrt_dect    =        hk_array[*, 57]  ; Vertical particles detected
    dead_time       = double(hk_array[*, 58]) ; dead time in second
    cpi_trig_set1   =        hk_array[*, 59]  ; CPI triggering settings
    cpi_trig_set2   =        hk_array[*, 60]  ; CPI triggering settings bits 11-16 min pixels before fire
    vrt_laser_drv   = double(hk_array[*, 61]) ; Vertical laser drive
    hrz_laser_drv   = double(hk_array[*, 62]) ; Horizontal laser drive
    vrt_masked_bits =        hk_array[*, 63]  ; # of vertical bits masked
    hrz_masked_bits =        hk_array[*, 64]  ; # of horizontal bits masked
    num_hrz_ovrld   =        hk_array[*, 65]  ; # of overload TWs found
    num_vrt_ovrld   =        hk_array[*, 66]  ; # of overload TWs found
    num_stereo_dect =        hk_array[*, 67]  ; # of overload TWs found
    comp_config     =        hk_array[*, 68]  ; Compression configuration

    pds_align_w1 = hk_array[*, 69] ; # of 4kb buffers sent last second
    pds_align_w2 = hk_array[*, 70] ; # of 4kb buffers sent last second

    TW_msw = hk_array[*, 71] ;  the three timing words
    TW_isw = hk_array[*, 72]
    TW_lsw = hk_array[*, 73]

    ;TW_vrt = hk_array[*, 72] * 2^16 + hk_array[*, 73] ; TAF this is only combining two timing words???

    ;    tasbin = dec2bin(data(HKDataIDX+74)*2^16 + data(HKDataIDX+75));%binary representation
    ; S=bin2dec(tasbin(1)); %floating point sign bit
    ;    E=bin2dec(tasbin(2:9));%floating point EEEEEEEE (exponential) field
    ;    Fraction = 1 + bin2dec(tasbin(10:32))/8388608

    ;    tas = [tas,(-1)^S*2^(E-127)*Fraction]

    TAS_array = two_uints2float(hk_array[*, 74], hk_array[*, 75])

    twods_cmd_acpt = hk_array[*, 76] ; % of 2ds commands accepted
    cpi_cmd_acpt = hk_array[*, 77] ; % of detected overlaps (from
    n2ds_data_blks = hk_array[*, 78] ; Number of 2D-S data blocks (4098 byte) sent over the last second
    array_skew_off = hk_array[*, 79] ; Array Skew Offset (six bits: 0-5)

    cpi_frame_rate = hk_array[*, 80] and '0FFF'X ; CPI Frame Rate


    ; combine the 3 components of the 48 bit timing word

    hk_probecount_array = ulong64(ishft(ulong64(TW_msw), 32) or ishft(ulong64(TW_isw), 16) or ulong64(TW_lsw))


    ; detect Hawykeye H 50 micron flag in HK Word 81, bit 15
    if((uint(hk_array[0, 79]) and uint('8000'X)) gt 0) then begin
        H50micron = 1
    endif else begin
        H50micron = 0
    endelse



    ; generate the output file, if needed
    ;    outfilename = 'read_hk_out_' + stripped_infilename + '.csv'

    ;    openw, olun, /get, outfilename

    ;    printf, olun, 'YYYY-MM-DD, HH:MM:SS.mmm, Hex Time, H Part, V Part, H Overflow, V Overflow, NULL Packets'

    ;    free_lun, olun


    ; generate some plots of the HK data

    ; also be good to look at the timing...

    if(show_hk_plots) then begin ; show_hk_plots toggled by keyword no_plots in call to read_hk
        if(plot_against_time) then begin
            ;            Result = JULDAY(Month, Day, Year, Hour, Minute, Second)
            ;            number_samples = 37
            ;            date_time = TIMEGEN(number_samples, UNITS = 'Seconds', $
            ;               START = JULDAY(3, 30, 2000, 14, 59, 30))
            ;            displacement = SIN(10.*!DTOR*FINDGEN(number_samples))

            date_labels = label_date(date_format='%H:%I:%S')

            jultimes = timegen(start=julday(hk_timestamp_array[0].month,          hk_timestamp_array[0].day,             hk_timestamp_array[0].year,             $
                hk_timestamp_array[0].hour,           hk_timestamp_array[0].minute,          hk_timestamp_array[0].second),          $
                final=julday(hk_timestamp_array[hk_count-1].month, hk_timestamp_array[hk_count-1].day,    hk_timestamp_array[hk_count-1].year,    $
                hk_timestamp_array[hk_count-1].hour,  hk_timestamp_array[hk_count-1].minute, hk_timestamp_array[hk_count-1].second), $
                units='Seconds')


            nxticks = 10

            window, 0, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            color = 5
            ;color_step = 8 ; 8 OK for 3V types, 27 plots * 8 = 216, into the reds
            color_step = 9 ; 27 plots * 9 = 243

            ;plot, temp_fst, xrange = [0, hk_count-1], /xst, yrange=[-60, 60], /yst, title='HK Temperatures', ytitle = 'Temperature (C)', xtitle = 'HK Packet'

            plot, jultimes, temp_fst, yrange=[-60, 60], /yst, title='HK Temperatures', ytitle = 'Temperature (C)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata
            ;xtickunits=['Time','Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, ymargin=[6,2] ; this label scheme has the major hours picked out on a lower line

            oplot, jultimes, temp_fst, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_uob, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_uob, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_lob, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_cst, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_ast, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_pyln1, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_pyln2, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_pyln3, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_ccd, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_il, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_ilz, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vlz, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hlz, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_pbd, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vpl, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vop, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hpl, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hop, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vmr, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hmr, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_ipl, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_dsp, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vayt, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_vayb, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hayt, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, temp_hayb, color = color
            color = (color + color_step) mod 255


            ; plot Relative Humidity, Can Pressure, and TAS
            window, 1, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 3] ; stack three plots vertically in window

            plot, jultimes, rh, yrange=[0, 100], /yst, title='HK Relative Humidity', ytitle = 'Relative Humidity (%)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata
            oplot, jultimes, rh, color = 170

            plot, jultimes, can_pres, yrange=[0, 15], /yst, title='HK Can Pressure', ytitle = 'Pressure (PSI)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata
            oplot, jultimes, can_pres, color = 70

            plot, jultimes, TAS_array, yrange=[0, 200], /yst, title='HK True Air Speed', ytitle = 'TAS (m/s)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata
            oplot, jultimes, TAS_array, color = 120

            !P.MULTI = stored_pmulti


            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 2] ; stack two plots vertically in window

            window, 2, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent


            plot, jultimes, v_ve0, yrange=[0, 5], /yst, title='HK Voltages (H diodes Reds, V diodes Greens)', ytitle = 'Voltage (V)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata

            ; plot V diodes in greens
            color = 170
            color_step = -11

            oplot, jultimes, v_ve0, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve21, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve42, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve64, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve85, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve106, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_ve127, color = color
            color = (color + color_step) mod 255

            ; plot H diodes in reds
            color = 254
            color_step = -8

            oplot, jultimes, v_he0, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he21, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he42, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he64, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he85, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he106, color = color
            color = (color + color_step) mod 255
            oplot, jultimes, v_he127, color = color
            color = (color + color_step) mod 255

            plot, jultimes, vrt_lsr, yrange=[0, 8], /yst, title='HK Voltages (Lasers and Raw+-)', ytitle = 'Voltage (V)', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata

            oplot, jultimes, vrt_lsr, color = 30
            oplot, jultimes, hrz_lsr, color = 70

            oplot, jultimes, raw_plusv,  color = 120
            oplot, jultimes, raw_minusv, color = 250

            !P.MULTI = stored_pmulti

            ;    can_pres

            ;    vrt_tec_i
            ;    hrz_tec_i

            ;    img_lsr_i



            window, 3, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            nmax = max(num_hrz_dect) > max(num_vrt_dect)
            if(nmax eq 0) then nmax = 2

            ; condition the data for the log plot -- don't output these values though

            temp_num_hrz_dect    = num_hrz_dect
            temp_num_vrt_dect    = num_vrt_dect
            temp_hrz_masked_bits = hrz_masked_bits
            temp_vrt_masked_bits = vrt_masked_bits
            temp_num_hrz_ovrld   = num_hrz_ovrld
            temp_num_vrt_ovrld   = num_vrt_ovrld

            index = where(temp_num_hrz_dect eq 0, count)
            if(count gt 0) then begin
                temp_num_hrz_dect[index] = 1
            endif

            index = where(temp_num_vrt_dect eq 0, count)
            if(count gt 0) then begin
                temp_num_vrt_dect[index] = 1
            endif

            index = where(temp_hrz_masked_bits eq 0, count)
            if(count gt 0) then begin
                temp_hrz_masked_bits[index] = 1
            endif

            index = where(temp_vrt_masked_bits eq 0, count)
            if(count gt 0) then begin
                temp_vrt_masked_bits[index] = 1
            endif

            index = where(temp_num_hrz_ovrld eq 0, count)
            if(count gt 0) then begin
                temp_num_hrz_ovrld[index] = 1
            endif

            index = where(temp_num_vrt_ovrld eq 0, count)
            if(count gt 0) then begin
                temp_num_vrt_ovrld[index] = 1
            endif

            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 2] ; stack two plots vertically in window

            plot, jultimes, temp_num_vrt_dect, yrange=[1, nmax], /yst, /ylog, title='HK Number Detected H (red) and V (green)', ytitle = 'Number of Particles', $
                xmargin=[10,10], $ ;, pos=[0.,.5,.9,1.]
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata

            if(total(num_hrz_dect) gt total(num_vrt_dect)) then begin
                oplot, jultimes, temp_num_hrz_dect,    color = 254
                oplot, jultimes, temp_num_vrt_dect,    color = 170;, linestyle=2
            endif else begin
                oplot, jultimes, temp_num_vrt_dect,    color = 170
                oplot, jultimes, temp_num_hrz_dect,    color = 254;, linestyle=2
            endelse

            nmax = max(hrz_masked_bits) > max(vrt_masked_bits) > max(num_hrz_ovrld) > max(num_vrt_ovrld)
            if(nmax eq 0) then nmax = 2

            plot, jultimes, temp_vrt_masked_bits, yrange=[0, nmax], ystyle=9, title='HK Overloads and Masks H (red) and V (green)', ytitle = 'Number of Masked Bits', $
                xmargin=[10,10], $ ;, pos=[0.,0.,.9,.5]
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata

            axis, color=0, yrange=[0, max(dead_time)], /yst, yaxis = 1, /save, ytitle='Dead Time (ms)' ; set up the right axis for dead time
            oplot, jultimes, dead_time, color = 70, linestyle=1

            axis, color=0, yrange=[0, nmax], /yst, yaxis = 0, /save ; reset to the left axis

            if(total(num_hrz_ovrld) gt total(num_vrt_ovrld)) then begin
                oplot, jultimes, temp_num_hrz_ovrld,   color = 254
                oplot, jultimes, temp_num_vrt_ovrld,   color = 170;, linestyle=2
            endif else begin
                oplot, jultimes, temp_num_vrt_ovrld,   color = 170
                oplot, jultimes, temp_num_hrz_ovrld,   color = 254;, linestyle=2
            endelse

            oplot, jultimes, temp_hrz_masked_bits, color = 220, linestyle=1
            oplot, jultimes, temp_vrt_masked_bits, color = 110, linestyle=1

            !P.MULTI = stored_pmulti



            ; plot the CPI trigger mode, Min Pixels before Fire, and CPI Frame Rate

            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 2] ; stack two plots vertically in window

            window, 4, xs = hk_plotsize_x, ys = hk_plotsize_y, xpos=300, ypos=300
            device, decomposed = 0
            loadct, 39, /silent


            ; CPI triggering settings bits 11-16 min pixels before fire  bits 6-7 CPI Trigger Source

            ; extract probe_mode, min pixels before fire, and

            probe_mode = ishft(probe_mode_raw, -6) and '0003'X

            cpi_min_pix = ishft(cpi_trig_set2, -11) and '03FF'X

            plot, jultimes, cpi_min_pix, title='CPI Modes, Standard vs Random (Green) and Min Pixels Before Fire (Red)', $
                xmargin=[10,10], $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                ytitle='Min Pixels Before Fire', yrange=[1, max(cpi_min_pix)+1], /yst, $
                background=255, color=0, /nodata

            oplot, jultimes, cpi_min_pix, color=254


            ; set up the right axis, and draw the axis labels in green to help identify the probe_mode line
            axis, color=0, yrange=[-1, 3], /yst, yaxis = 1, /save, yticks=4, ytickname=[' ', 'Standard', ' ', 'Random', ' ']
            axis, yrange=[-1, 3], /yst, yaxis = 1, /save, yticks=4, ytickname=[' ', 'Standard', ' ', 'Random', ' '], yticklayout=1, color=170

            oplot, jultimes, probe_mode, color=170


            ; cpi_frame_rate = hk_array[*, 80] and '0FFF'X ; CPI Frame Rate

            plot, jultimes, cpi_frame_rate, title='CPI Frame Rate', $
                xmargin=[10,10], ytitle='Frames/s', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                background=255, color=0, /nodata

            oplot, jultimes, cpi_frame_rate, color=130


            !P.MULTI = stored_pmulti

        endif else begin ; plot against time (above) vs plot against HK number (below)

            window, 0, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            color = 5
            color_step = 8

            plot, temp_fst, xrange = [0, hk_count-1], /xst, yrange=[-60, 60], /yst, title='HK Temperatures', ytitle = 'Temperature (C)', xtitle = 'HK Packet'
            oplot, temp_fst, color = color
            color = (color + color_step) mod 255
            oplot, temp_uob, color = color
            color = (color + color_step) mod 255
            oplot, temp_uob, color = color
            color = (color + color_step) mod 255
            oplot, temp_lob, color = color
            color = (color + color_step) mod 255
            oplot, temp_cst, color = color
            color = (color + color_step) mod 255
            oplot, temp_ast, color = color
            color = (color + color_step) mod 255
            oplot, temp_pyln1, color = color
            color = (color + color_step) mod 255
            oplot, temp_pyln2, color = color
            color = (color + color_step) mod 255
            oplot, temp_pyln3, color = color
            color = (color + color_step) mod 255
            oplot, temp_ccd, color = color
            color = (color + color_step) mod 255
            oplot, temp_il, color = color
            color = (color + color_step) mod 255
            oplot, temp_ilz, color = color
            color = (color + color_step) mod 255
            oplot, temp_vlz, color = color
            color = (color + color_step) mod 255
            oplot, temp_hlz, color = color
            color = (color + color_step) mod 255
            oplot, temp_pbd, color = color
            color = (color + color_step) mod 255
            oplot, temp_vpl, color = color
            color = (color + color_step) mod 255
            oplot, temp_vop, color = color
            color = (color + color_step) mod 255
            oplot, temp_hpl, color = color
            color = (color + color_step) mod 255
            oplot, temp_hop, color = color
            color = (color + color_step) mod 255
            oplot, temp_vmr, color = color
            color = (color + color_step) mod 255
            oplot, temp_hmr, color = color
            color = (color + color_step) mod 255
            oplot, temp_ipl, color = color
            color = (color + color_step) mod 255
            oplot, temp_dsp, color = color
            color = (color + color_step) mod 255
            oplot, temp_vayt, color = color
            color = (color + color_step) mod 255
            oplot, temp_vayb, color = color
            color = (color + color_step) mod 255
            oplot, temp_hayt, color = color
            color = (color + color_step) mod 255
            oplot, temp_hayb, color = color
            color = (color + color_step) mod 255


            ; plot Relative Humidity, Can Pressure, and TAS
            window, 1, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 3] ; stack three plots vertically in window

            plot, rh, xrange = [0, hk_count-1], /xst, yrange=[0, 100], /yst, title='HK Relative Humidity', ytitle = 'Relative Humidity (%)', xtitle = 'HK Packet'
            oplot, rh, color = 170

            pressure_range_max = 15 < ceil(max(can_pres)) ; normally use 15 PSI, or use the ceiling of the max can_pressure if higher (sometimes a sealed hot can can exceed 15 PSI)
            plot, can_pres, xrange = [0, hk_count-1], /xst, yrange=[0, pressure_range_max], /yst, title='HK Can Pressure', ytitle = 'Pressure (PSI)', xtitle = 'HK Packet'
            oplot, can_pres, color = 70

            plot, TAS_array, xrange = [0, hk_count-1], /xst, yrange=[0, 200], /yst, title='HK True Air Speed', ytitle = 'TAS (m/s)', xtitle = 'HK Packet'
            oplot, TAS_array, color = 120

            !P.MULTI = stored_pmulti


            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 2] ; stack two plots vertically in window

            window, 2, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent


            plot, v_ve0, xrange = [0, hk_count-1], /xst, yrange=[0, 5], /yst, title='HK Voltages (H diodes Reds, V diodes Greens)', ytitle = 'Voltage (V)', xtitle = 'HK Packet'

            ; plot V diodes in greens
            color = 170
            color_step = -11

            oplot, v_ve0, color = color
            color = (color + color_step) mod 255
            oplot, v_ve21, color = color
            color = (color + color_step) mod 255
            oplot, v_ve42, color = color
            color = (color + color_step) mod 255
            oplot, v_ve64, color = color
            color = (color + color_step) mod 255
            oplot, v_ve85, color = color
            color = (color + color_step) mod 255
            oplot, v_ve106, color = color
            color = (color + color_step) mod 255
            oplot, v_ve127, color = color
            color = (color + color_step) mod 255

            ; plot H diodes in reds
            color = 254
            color_step = -8

            oplot, v_he0, color = color
            color = (color + color_step) mod 255
            oplot, v_he21, color = color
            color = (color + color_step) mod 255
            oplot, v_he42, color = color
            color = (color + color_step) mod 255
            oplot, v_he64, color = color
            color = (color + color_step) mod 255
            oplot, v_he85, color = color
            color = (color + color_step) mod 255
            oplot, v_he106, color = color
            color = (color + color_step) mod 255
            oplot, v_he127, color = color
            color = (color + color_step) mod 255

            plot, vrt_lsr, xrange = [0, hk_count-1], /xst, yrange=[0, 8], /yst, title='HK Voltages (Lasers and Raw+-)', ytitle = 'Voltage (V)', xtitle = 'HK Packet'

            oplot, vrt_lsr, color = 30
            oplot, hrz_lsr, color = 70

            oplot, raw_plusv,  color = 120
            oplot, raw_minusv, color = 250

            !P.MULTI = stored_pmulti

            ;    can_pres

            ;    vrt_tec_i
            ;    hrz_tec_i

            ;    img_lsr_i



            window, 3, xs = hk_plotsize_x, ys = hk_plotsize_y
            device, decomposed = 0
            loadct, 39, /silent

            nmax = max(num_hrz_dect) > max(num_vrt_dect)
            if(nmax eq 0) then nmax = 2

            ; condition the data for the log plot -- don't output these values though

            temp_num_hrz_dect    = num_hrz_dect
            temp_num_vrt_dect    = num_vrt_dect
            temp_hrz_masked_bits = hrz_masked_bits
            temp_vrt_masked_bits = vrt_masked_bits
            temp_num_hrz_ovrld   = num_hrz_ovrld
            temp_num_vrt_ovrld   = num_vrt_ovrld

            index = where(temp_num_hrz_dect eq 0, count)
            if(count gt 0) then begin
                temp_num_hrz_dect[index] = 1
            endif

            index = where(temp_num_vrt_dect eq 0, count)
            if(count gt 0) then begin
                temp_num_vrt_dect[index] = 1
            endif

            index = where(temp_hrz_masked_bits eq 0, count)
            if(count gt 0) then begin
                temp_hrz_masked_bits[index] = 1
            endif

            index = where(temp_vrt_masked_bits eq 0, count)
            if(count gt 0) then begin
                temp_vrt_masked_bits[index] = 1
            endif

            index = where(temp_num_hrz_ovrld eq 0, count)
            if(count gt 0) then begin
                temp_num_hrz_ovrld[index] = 1
            endif

            index = where(temp_num_vrt_ovrld eq 0, count)
            if(count gt 0) then begin
                temp_num_vrt_ovrld[index] = 1
            endif

            stored_pmulti = !P.MULTI
            !P.MULTI = [0, 1, 2] ; stack two plots vertically in window

            plot, temp_num_hrz_dect, xrange = [0, hk_count-1], /xst, yrange=[1, nmax], /yst, /ylog, title='HK Number Detected H (red) and V (green)', ytitle = 'Number of Particles', xtitle = 'HK Packet', xmargin=[10,10] ;, pos=[0.,.5,.9,1.]

            if(total(num_hrz_dect) gt total(num_vrt_dect)) then begin
                oplot, temp_num_hrz_dect,    color = 254
                oplot, temp_num_vrt_dect,    color = 170;, linestyle=2
            endif else begin
                oplot, temp_num_vrt_dect,    color = 170
                oplot, temp_num_hrz_dect,    color = 254;, linestyle=2
            endelse

            nmax = max(hrz_masked_bits) > max(vrt_masked_bits) > max(num_hrz_ovrld) > max(num_vrt_ovrld)
            if(nmax eq 0) then nmax = 2

            plot, temp_hrz_masked_bits, xrange = [0, hk_count-1], /xst, yrange=[0, nmax], ystyle=9, title='HK Overloads and Masks H (red) and V (green)', ytitle = 'Number of Masked Bits', xtitle = 'HK Packet', xmargin=[10,10] ;, pos=[0.,0.,.9,.5]

            axis, yrange=[0, max(dead_time)], /yst, yaxis = 1, /save, ytitle='Dead Time (ms)' ; set up the right axis for dead time
            oplot, dead_time, color = 70, linestyle=1

            axis, yrange=[0, nmax], /yst, yaxis = 0, /save ; reset to the left axis

            if(total(num_hrz_ovrld) gt total(num_vrt_ovrld)) then begin
                oplot, temp_num_hrz_ovrld,   color = 254
                oplot, temp_num_vrt_ovrld,   color = 170;, linestyle=2
            endif else begin
                oplot, temp_num_vrt_ovrld,   color = 170
                oplot, temp_num_hrz_ovrld,   color = 254;, linestyle=2
            endelse

            oplot, temp_hrz_masked_bits, color = 220, linestyle=1
            oplot, temp_vrt_masked_bits, color = 110, linestyle=1

            !P.MULTI = stored_pmulti


            ; CPI triggering settings bits 11-16 min pixels before fire  bits 6-7 CPI Trigger Source

            ; extract probe_mode, min pixels before fire, and

            probe_mode = ishft(probe_mode_raw, -6) and '0003'X

            cpi_min_pix = ishft(cpi_trig_set2, -11) and '03FF'X

            plot, jultimes, cpi_min_pix, title='CPI Modes, Standard vs Random (Green) and Min Pixels Before Fire (Red)', $
                xmargin=[10,10], $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks, $
                ytitle='Min Pixels Before Fire', yrange=[1, max(cpi_min_pix)+1], /yst

            oplot, jultimes, cpi_min_pix, color=254


            ; set up the right axis, and draw the axis labels in green to help identify the probe_mode line
            axis, yrange=[-1, 3], /yst, yaxis = 1, /save, yticks=4, ytickname=[' ', 'Standard', ' ', 'Random', ' ']
            axis, yrange=[-1, 3], /yst, yaxis = 1, /save, yticks=4, ytickname=[' ', 'Standard', ' ', 'Random', ' '], yticklayout=1, color=170

            oplot, jultimes, probe_mode, color=170


            ; cpi_frame_rate = hk_array[*, 80] and '0FFF'X ; CPI Frame Rate

            plot, jultimes, cpi_frame_rate, title='CPI Frame Rate', $
                xmargin=[10,10], ytitle='Frames/s', $
                xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks

            oplot, jultimes, cpi_frame_rate, color=130


            !P.MULTI = stored_pmulti

        endelse ; plot_against_time


        ;window, 5, xs = hk_plotsize_x, ys = hk_plotsize_y
        ;device, decomposed = 0
        ;loadct, 39, /silent

        ;    dead_time

        ;    vrt_laser_drv
        ;    hrz_laser_drv

        ;    TW_msw
        ;    TW_isw
        ;    TW_lsw


        if(save_hk_plots) then begin
            ; set up the plot file names

            date_string = string(format='(%"%04d-%02d-%02d_%02d.%02d.%02d.%03d")', $
                hk_timestamp_array[0].year,   $
                hk_timestamp_array[0].month,  $
                hk_timestamp_array[0].day,    $
                hk_timestamp_array[0].hour,   $
                hk_timestamp_array[0].minute, $
                hk_timestamp_array[0].second, $
                hk_timestamp_array[0].ms)

            ;file_extension = '.jpg'
            file_extension = '.png'

            outfilename = infilepath + 'batch2d_output\HK_temp_'   + date_string + file_extension
            wset, 0
            ;jpger, outfilename
            pnger, outfilename

            outfilename = infilepath + 'batch2d_output\HK_env_'    + date_string + file_extension
            wset, 1
            ;jpger, outfilename
            pnger, outfilename

            outfilename = infilepath + 'batch2d_output\HK_volt_'   + date_string + file_extension
            wset, 2
            ;jpger, outfilename
            pnger, outfilename

            outfilename = infilepath + 'batch2d_output\HK_counts_' + date_string + file_extension
            wset, 3
            ;jpger, outfilename
            pnger, outfilename

            outfilename = infilepath + 'batch2d_output\CPI_modes_' + date_string + file_extension
            wset, 4
            ;jpger, outfilename
            pnger, outfilename

        endif

        if(!output_CPI_frames_csv) then begin
            ; output csv file of CPI frame rate and min pixel mode, with the header:
            ; Time UTC, CPI Frame Rate, Min Pixels Before Fire

            date_string = string(format='(%"%04d-%02d-%02d_%02d.%02d.%02d.%03d")', $
                hk_timestamp_array[0].year,   $
                hk_timestamp_array[0].month,  $
                hk_timestamp_array[0].day,    $
                hk_timestamp_array[0].hour,   $
                hk_timestamp_array[0].minute, $
                hk_timestamp_array[0].second, $
                hk_timestamp_array[0].ms)

            outfilename = infilepath + 'HK_CPI_frames_' + date_string + '.csv'

            openw,  olunCPI, /get, outfilename
            printf, olunCPI, 'Time UTC, CPI Frame Rate, Min Pixels Before Fire'

            for i=0L, hk_count-1 do begin
                printf, olunCPI, format='(%"%02d:%02d:%02d, %5d, %3d")', $
                    hk_timestamp_array[i].hour, $
                    hk_timestamp_array[i].minute, $
                    hk_timestamp_array[i].second, $
                    cpi_frame_rate[i], $
                    cpi_min_pix[i]

            endfor

            free_lun, olunCPI
        endif

        if(!output_TAS_file) then begin
            ; output csv file of UTC seconds and TAS
            ; Time UTC (seconds), TAS (m/s)

            date_string = string(format='(%"%04d-%02d-%02d_%02d.%02d.%02d.%03d")', $
                hk_timestamp_array[0].year,   $
                hk_timestamp_array[0].month,  $
                hk_timestamp_array[0].day,    $
                hk_timestamp_array[0].hour,   $
                hk_timestamp_array[0].minute, $
                hk_timestamp_array[0].second, $
                hk_timestamp_array[0].ms)

            outfilename = infilepath + 'batch2d_output\TAS_3V2DS_' + date_string + '.txt'

            openw,  olunTAS, /get, outfilename

            ;printf, olunTAS, 'Time UTC (seconds), TAS (m/s)'

            for i=0L, hk_count-1 do begin
                printf, olunTAS, format='(%"%6d, %6.1f")', $
                    long(3600.*hk_timestamp_array[i].hour + 60.*hk_timestamp_array[i].minute + long(hk_timestamp_array[i].second)), $
                    TAS_array[i]

            endfor

            free_lun, olunTAS
        endif
    endif ; show_plots





    ; output csv file of HK data
    if(!output_HK_csv) then begin
        ; Output the probe-specific CSV header and HK data
        ; Handled here are the 3V type probes, the 3V-CPI, and uCOPP, for 2D type probes, see read_hk_2d.pro

        ;defsysv, '!MODE_2D128',       1
        ;defsysv, '!MODE_2DS',         2
        ;defsysv, '!MODE_3VCPI',       3
        ;defsysv, '!MODE_HVPS',        4
        ;defsysv, '!MODE_uCOPP',       5

        outfilename = infilepath + 'batch2d_output\' + stripped_infilename + '_HK.csv'

        openw,  oluncsv, /get, outfilename



        if(!probe_type eq !MODE_3VCPI) then begin
            printf, oluncsv, $
                'Date,Time,sod,#H,#V,Hoverload,Voverload,Hmask,Vmask,Hlaser,Vlaser,TAS,2DSCmds,CPICmds,PDSinfo1,PDSinfo2,N2DSblks,Nstereo,SkewOff,Compression,ModeRaw,CPIFrameRate,' + $
                'ImgLsrCur,ImgLsrPwr,ImgLiSet,ImgLpSet,HeatStatus,UpOptBT,LoOptBT,FwdTubeT,CenTubeT,AftTubeT,Pylon1T,Pylon2T,Pylon3T,CCDT,ImgLensT,ImgLaserT,' + $
                'HlaserT,VlaserT,PowerT,HplatenT,VplatenT,HopticsT,VopticsT,HmirrorT,VmirrorT,IntPlatenT,DSPT,HarrTopT,VarrTopT,HarrBotT,VarrBotT,' + $
                'RH%,Pressure,Htec,Vtec,+7V,-7V,H0V,H21V,H42V,H64V,H85V,H106V,H127V,V0V,V21V,V42V,V64V,V85V,V106V,V127V,probecount,' + string(format='(%"H50Micron = %d")', H50Micron) ; 50 micron indicator just in header, not a column

            for i=0L, hk_count-1 do begin

                date_time_string = string(format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d, %6d")', $
                    hk_timestamp_array[i].year,   $
                    hk_timestamp_array[i].month,  $
                    hk_timestamp_array[i].day,    $
                    hk_timestamp_array[i].hour,   $
                    hk_timestamp_array[i].minute, $
                    hk_timestamp_array[i].second, $
                    hk_timestamp_array[i].ms,    $
                    long(3600.*hk_timestamp_array[i].hour + 60.*hk_timestamp_array[i].minute + long(hk_timestamp_array[i].second)))

                if(comp_config[i] eq 0) then begin
                    compression_string = 'STEREO'
                endif else if(comp_config[i] eq 1) then begin
                    compression_string = 'Both'
                endif else if(comp_config[i] eq 2) then begin
                    compression_string = 'H Only'
                endif else if(comp_config[i] eq 3) then begin
                    compression_string = 'V Only'
                endif else begin
                    compression_string = 'ERROR'
                endelse

                outstring1 = string($
                    format='(%"%s,%d,%d,%d,%d,%d,%d,%d,%d,%5.1f,%d,%d,%d,%d,%d,%d,%d,%s,0x%04X,%d,%5.1f,%5.1f,%5.1f,%5.1f,0x%04X,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,")',  $
                    date_time_string, $
                    num_hrz_dect[i], $
                    num_vrt_dect[i], $
                    num_hrz_ovrld[i], $
                    num_vrt_ovrld[i], $
                    hrz_masked_bits[i], $
                    vrt_masked_bits[i], $
                    hrz_laser_drv[i], $
                    vrt_laser_drv[i], $
                    TAS_array[i], $
                    twods_cmd_acpt[i], $
                    cpi_cmd_acpt[i], $
                    pds_align_w1[i], $
                    pds_align_w2[i], $
                    n2ds_data_blks[i], $
                    num_stereo_dect[i], $
                    array_skew_off[i], $
                    compression_string, $
                    probe_mode_raw[i], $
                    cpi_frame_rate[i], $
                    img_lsr_i[i], $
                    img_lsr_pw[i], $
                    img_lsr_i_set[i], $
                    img_lsr_pw_set[i], $
                    heater_status[i], $
                    temp_uob[i], $
                    temp_lob[i], $
                    temp_fst[i], $
                    temp_cst[i], $
                    temp_ast[i], $
                    temp_pyln1[i], $
                    temp_pyln2[i], $
                    temp_pyln3[i], $
                    temp_ccd[i], $
                    temp_il[i], $
                    temp_ilz[i])

                outstring2 = string($
                    format='(%"%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,0x%012X")',  $
                    temp_hlz[i], $
                    temp_vlz[i], $
                    temp_pbd[i], $
                    temp_hpl[i], $
                    temp_vpl[i], $
                    temp_hop[i], $
                    temp_vop[i], $
                    temp_hmr[i], $
                    temp_vmr[i], $
                    temp_ipl[i], $
                    temp_dsp[i], $
                    temp_hayt[i], $
                    temp_vayt[i], $
                    temp_hayb[i], $
                    temp_vayb[i], $
                    rh[i], $
                    can_pres[i], $
                    hrz_tec_i[i], $
                    vrt_tec_i[i], $
                    raw_plusv[i], $
                    -raw_minusv[i], $
                    v_he0[i], $
                    v_he21[i], $
                    v_he42[i], $
                    v_he64[i], $
                    v_he85[i], $
                    v_he106[i], $
                    v_he127[i], $
                    v_ve0[i], $
                    v_ve21[i], $
                    v_ve42[i], $
                    v_ve64[i], $
                    v_ve85[i], $
                    v_ve106[i], $
                    v_ve127[i], $
                    hk_probecount_array[i])

                outstring = string(format='(%"%s%s")', outstring1, outstring2) ; format string was too long for IDL, had to split format

                printf, oluncsv, outstring
            endfor
        endif else if(!probe_type eq !MODE_uCOPP) then begin ; NOTE: untested uCOPP -- !flag_uCOPP didn't seem to work, need to check
            printf, oluncsv, $
                'Date,Time,sod,#V,Voverload,Vmask,Vlaser,TAS,2DSCmds,CPICmds,PDSinfo1,PDSinfo2,N2DSblks,Nstereo,SkewOff,Compression,ModeRaw,CPIFrameRate,' + $
                'ImgLsrCur,ImgLsrPwr,ImgLiSet,ImgLpSet,HeatStatus,UpOptBT,LoOptBT,FwdTubeT,CenTubeT,AftTubeT,Pylon1T,Pylon2T,Pylon3T,CCDT,ImgLensT,ImgLaserT,' + $
                'VlaserT,PowerT,VplatenT,VopticsT,VmirrorT,IntPlatenT,DSPT,VarrTopT,VarrBotT,' + $
                'RH%,Pressure,Vtec,+7V,-7V,V0V,V21V,V42V,V64V,V85V,V106V,V127V,probecount,' + string(format='(%"H50Micron = %d")', H50Micron) ; 50 micron indicator just in header, not a column

            for i=0L, hk_count-1 do begin

                date_time_string = string(format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d, %6d")', $
                    hk_timestamp_array[i].year,   $
                    hk_timestamp_array[i].month,  $
                    hk_timestamp_array[i].day,    $
                    hk_timestamp_array[i].hour,   $
                    hk_timestamp_array[i].minute, $
                    hk_timestamp_array[i].second, $
                    hk_timestamp_array[i].ms,    $
                    long(3600.*hk_timestamp_array[i].hour + 60.*hk_timestamp_array[i].minute + long(hk_timestamp_array[i].second)))

                if(comp_config[i] eq 0) then begin
                    compression_string = 'STEREO'
                endif else if(comp_config[i] eq 1) then begin
                    compression_string = 'Both'
                endif else if(comp_config[i] eq 2) then begin
                    compression_string = 'H Only'
                endif else if(comp_config[i] eq 3) then begin
                    compression_string = 'V Only'
                endif else begin
                    compression_string = 'ERROR'
                endelse

                outstring1 = string($
                    format='(%"%s,%d,%d,%d,%d,%5.1f,%d,%d,%d,%d,%d,%d,%d,%s,0x%04X,%d,%5.1f,%5.1f,%5.1f,%5.1f,0x%04X,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,")',  $
                    date_time_string, $
                    num_vrt_dect[i], $
                    num_vrt_ovrld[i], $
                    vrt_masked_bits[i], $
                    vrt_laser_drv[i], $
                    TAS_array[i], $
                    twods_cmd_acpt[i], $
                    cpi_cmd_acpt[i], $
                    pds_align_w1[i], $
                    pds_align_w2[i], $
                    n2ds_data_blks[i], $
                    num_stereo_dect[i], $
                    array_skew_off[i], $
                    compression_string, $
                    probe_mode_raw[i], $
                    cpi_frame_rate[i], $
                    img_lsr_i[i], $
                    img_lsr_pw[i], $
                    img_lsr_i_set[i], $
                    img_lsr_pw_set[i], $
                    heater_status[i], $
                    temp_uob[i], $
                    temp_lob[i], $
                    temp_fst[i], $
                    temp_cst[i], $
                    temp_ast[i], $
                    temp_pyln1[i], $
                    temp_pyln2[i], $
                    temp_pyln3[i], $
                    temp_ccd[i], $
                    temp_il[i], $
                    temp_ilz[i])

                outstring2 = string($
                    format='(%"%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,%5.2f,0x%012X")',  $
                    temp_vlz[i], $
                    temp_pbd[i], $
                    temp_vpl[i], $
                    temp_vop[i], $
                    temp_vmr[i], $
                    temp_ipl[i], $
                    temp_dsp[i], $
                    temp_vayt[i], $
                    temp_vayb[i], $
                    rh[i], $
                    can_pres[i], $
                    vrt_tec_i[i], $
                    raw_plusv[i], $
                    -raw_minusv[i], $
                    v_ve0[i], $
                    v_ve21[i], $
                    v_ve42[i], $
                    v_ve64[i], $
                    v_ve85[i], $
                    v_ve106[i], $
                    v_ve127[i], $
                    hk_probecount_array[i])

                outstring = string(format='(%"%s%s")', outstring1, outstring2) ; format string was too long for IDL, had to split format

                printf, oluncsv, outstring
            endfor
        endif ; probe specific HK

        free_lun, oluncsv
    endif ; print HK csv






    ; populate HK1 and HK2

    HK1_timestamp.year    = hk_timestamp_array[0].year
    HK1_timestamp.month   = hk_timestamp_array[0].month
    HK1_timestamp.day     = hk_timestamp_array[0].day
    HK1_timestamp.hour    = hk_timestamp_array[0].hour
    HK1_timestamp.minute  = hk_timestamp_array[0].minute
    HK1_timestamp.second  = hk_timestamp_array[0].second
    HK1_timestamp.ms      = hk_timestamp_array[0].ms
    HK1_timestamp.us      = 0
    HK1_timestamp.ns      = 0
    HK1_timestamp.counter = hk_probecount_array[0]

    HK2_timestamp.year    = hk_timestamp_array[1].year
    HK2_timestamp.month   = hk_timestamp_array[1].month
    HK2_timestamp.day     = hk_timestamp_array[1].day
    HK2_timestamp.hour    = hk_timestamp_array[1].hour
    HK2_timestamp.minute  = hk_timestamp_array[1].minute
    HK2_timestamp.second  = hk_timestamp_array[1].second
    HK2_timestamp.ms      = hk_timestamp_array[1].ms
    HK2_timestamp.us      = 0
    HK2_timestamp.ns      = 0
    HK2_timestamp.counter = hk_probecount_array[1]


    HK_timediff_nanos = calc_timediff_nanos(HK1_timestamp, HK2_timestamp)


    ;stop




    ;stop

end ; read_hk_3v
